# Copyright (c) 2014 Baidu.com, Inc. All Rights Reserved
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions
# and limitations under the License.

"""
This module provides a client class for ACL.
"""

import copy
import json
import logging
import uuid
import sys

from baidubce import bce_base_client
from baidubce.auth import bce_v1_signer
from baidubce.http import bce_http_client
from baidubce.http import handler
from baidubce.http import http_methods
from baidubce import utils
from baidubce.utils import required
from baidubce import compat

if sys.version < '3':
    sys.setdefaultencoding('utf-8')

_logger = logging.getLogger(__name__)


class AclClient(bce_base_client.BceBaseClient):
    """
    ACL base sdk client
    """
    prefix = b'/v1'

    def __init__(self, config=None):
        bce_base_client.BceBaseClient.__init__(self, config)

    def _merge_config(self, config=None):
        """
        :param config:
        :type config: baidubce.BceClientConfiguration
        :return:
        """
        if config is None:
            return self.config
        else:
            new_config = copy.copy(self.config)
            new_config.merge_non_none_values(config)
            return new_config

    def _send_request(self, http_method, path,
                      body=None, headers=None, params=None,
                      config=None, body_parser=None):
        config = self._merge_config(config)
        if body_parser is None:
            body_parser = handler.parse_json
        if headers is None:
            headers = {b'Accept': b'*/*', b'Content-Type':
                b'application/json;charset=utf-8'}

        return bce_http_client.send_request(
            config, bce_v1_signer.sign, [handler.parse_error, body_parser],
            http_method, path, body, headers, params)

    @required(vpc_id=(bytes, str))
    def list_acl_entrys(self, vpc_id, config=None):
        """
        Get the detail information of acl for specific vpc.

        :param vpc_id:
            the vpc id
        :type vpc_id: string

        :param config:
        :type config: baidubce.BceClientConfiguration

        :return:
        :rtype baidubce.bce_response.BceResponse
        """
        path = utils.append_uri(self.prefix, 'acl')
        params = {}
        params[b'vpcId'] = vpc_id

        return self._send_request(http_methods.GET, path,
                                  params=params, config=config)

    @required(rule_list=list)
    def create_acl(self, rule_list, client_token=None, config=None):
        """
        Create  acl rules with the specified options.

        :param rule_list:
                a list contains acl rules.
                https://cloud.baidu.com/doc/VPC/API.html#AclRuleRequest
                The elements of the list are AclRuleRequest
        :type rule_list: list

        AclRuleRequest{
            :param protocol:
                The parameter specify which protocol will the acl rule work on
            :value: "all" or ""tcp" or "udp" or "icmp"
            :type protocol: string

            :param sourceIpAddress:
                Source ip address which the rule applied to
            :type sourceIpAddress: string

            :param destinationIpAddress:
                Destination ip address which the rule applied to
            :type destinationIpAddress: string

            :param sourcePort:
                Port used by source ip address
            :value 1-65535
            :type sourcePort: string

            :param destinationPort:
                Port used by destination ip address
            :value 1-65535
            :type destinationPort:string

            :param position:
                Priority of the rule
            :value 1-5000,unique,The smaller the value, the higher the priority
            :type:position:Integer

            :param direction:
                The rule is a ingress or a egress rule
            :value: "ingress" or "egress"
            :type direction:string

            :param action:
                The rule is allowed or denied
            :value "allow" or "deny"
            :type action:string

            :param description(Optional):
                The option param to describe the acl rule.
            :type description: string
        }

        :param client_token:
            If the clientToken is not specified by the user,
            a random Stringgenerated by default algorithm will be used.
        :type client_token: string

        :param config:
        :type config: baidubce.BceClientConfiguration

        :return:
        :rtype baidubce.bce_response.BceResponse
        """
        path = utils.append_uri(self.prefix, 'acl', 'rule')
        params = {}

        if client_token is None:
            params[b'clientToken'] = generate_client_token()
        else:
            params[b'clientToken'] = client_token

        body = {
            'aclRules': rule_list
        }

        return self._send_request(http_methods.POST, path,
                                  body=json.dumps(body), params=params,
                                  config=config)

    @required(subnet_id=(bytes, str))
    def list_subnet_acl(self, subnet_id, marker=None, max_keys=None, config=None):
        """
        Return a list of acl rules of specify subnet.

        :param subnet_id
            the id of subnet whhich the acl applied
        :type subnet_id: string

        :param marker
            The optional parameter marker specified in the original
            request to specify where in the results to begin listing.
            Together with the marker, specifies the list result
            which listing should begin. If the marker is not specified,
            the list result will listing from the first one.
        :type marker: string

        :param max_keys
            The optional parameter to specifies the max number of
            list result to return.
            The default value is 1000.
        :type max_keys: int

        :param config:
        :type config: baidubce.BceClientConfiguration

        :return:
        :rtype baidubce.bce_response.BceResponse
         """
        path = utils.append_uri(self.prefix, 'acl', 'rule')
        params = {}

        if marker is not None:
            params[b'marker'] = marker
        if max_keys is not None:
            params[b'maxKeys'] = max_keys

        params[b'subnetId'] = subnet_id
        return self._send_request(http_methods.GET, path,
                                  params=params, config=config)

    @required(acl_rule_id=(bytes, str))
    def delete_acl(self, acl_rule_id, client_token=None, config=None):
        """
        Delete the  specific acl rule.

        :param acl_rule_id:
            The id of the specified acl.
        :type acl_rule_id: string

        :param client_token:
            If the clientToken is not specified by the user, a random String
            generated by default algorithm will be used.
        :type client_token: string

        :param config:
        :type config: baidubce.BceClientConfiguration

        :return:
        :rtype baidubce.bce_response.BceResponse
        """
        path = utils.append_uri(self.prefix, 'acl', 'rule', acl_rule_id)
        params = {}
        if client_token is None:
            params[b'clientToken'] = generate_client_token()
        else:
            params[b'clientToken'] = client_token
        return self._send_request(http_methods.DELETE, path,
                                  params=params, config=config)

    @required(acl_rule_id=(bytes, str))
    def update_acl(self, acl_rule_id, description=None,
                   protocol=None, source_ip_address=None,
                   destination_ip_address=None, source_port=None,
                   destination_port=None,
                   position=None, action=None,
                   client_token=None, config=None):
        """
        Modify the special attribute to new value of the acl owned by the user.

        :param acl_rule_id
                id of the acl to be modified
        :type acl_rule_id:string

        :param description:
                The option param to describe the acl rule.
        :type description: string

        :param protocol:
                 The parameter specify which protocol will the acl rule work on
        :value: "all" or ""tcp" or "udp" or "icmp"
        :type protocol: string

        :param source_ip_address:
                 source ip address which the rule applied to
        :type source_ip_address: string

        :param destination_ip_address:
                 destination ip address which the rule applied to
        :type destination_ip_address: string

        :param source_port:
                 port used by source ip address
        :value 1-65535
        :type source_port: string

        :param destination_port:
                 port used by destination ip address
        :value 1-65535
        :type destination_port:string

        :param position:
                 priority of the rule
        :value 1-5000,unique,The smaller the value, the higher the priority
        :type:position:Integer

        :param action:
                the rule is allowed or denied
        :value "allow" or "deny"
        :type action:string

        :param client_token:
                If the clientToken is not specified by the user, a random
                String generated by default algorithm will be used.
        :type client_token: string

        :param config:
        :type config: baidubce.BceClientConfiguration

        :return:
        :rtype baidubce.bce_response.BceResponse
        """
        path = utils.append_uri(self.prefix, 'acl', 'rule', acl_rule_id)
        params = {}

        if client_token is None:
            params[b'clientToken'] = generate_client_token()
        else:
            params[b'clientToken'] = client_token

        body = {}
        if description is not None:
            body['description'] = compat.convert_to_string(description)

        if protocol is not None:
            body['protocol'] = compat.convert_to_string(protocol)

        if source_ip_address is not None:
            body['sourceIpAddress'] = \
                compat.convert_to_string(source_ip_address)

        if destination_ip_address is not None:
            body['destinationIpAddress'] = \
                compat.convert_to_string(destination_ip_address)

        if source_port is not None:
            body['sourcePort'] = compat.convert_to_string(source_port)

        if destination_port is not None:
            body['destinationPort'] = \
                compat.convert_to_string(destination_port)

        if position is not None:
            body['position'] = position

        if action is not None:
            body['action'] = compat.convert_to_string(action)

        return self._send_request(http_methods.PUT, path, json.dumps(body),
                                  params=params, config=config)


def generate_client_token_by_uuid():
    """
    The default method to generate the random string for client_token
    if the optional parameter client_token is not specified by the user.

    :return:
    :rtype string
    """
    return str(uuid.uuid4())


generate_client_token = generate_client_token_by_uuid



# Copyright 2014 Baidu, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file
# except in compliance with the License. You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the
# License is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND,
# either express or implied. See the License for the specific language governing permissions
# and limitations under the License.
"""
This module provides models for BES-SDK.
"""


class Billing(object):
    """
    Billing Class
    """

    def __init__(self, payment_type, time):
        self.paymentType = payment_type
        self.time = time


class Module(object):
    """
    Module Class
    """

    def __init__(self, type=None,
                 instance_num=None,
                 version=None,
                 slot_type=None,
                 desire_instance_num=None,
                 disk_slot_info=None):
        if type is not None:
            self.type = type
        if instance_num is not None:
            self.instanceNum = instance_num
        if version is not None:
            self.version = version
        if slot_type is not None:
            self.slotType = slot_type
        if desire_instance_num is not None:
            self.desireInstanceNum = desire_instance_num
        if disk_slot_info is not None:
            self.diskSlotInfo = disk_slot_info.__dict__


class DiskSlotInfo(object):
    """
    DiskSlotInfo class
    """

    def __init__(self, type=None, size=None, instance_num=None):
        if type is not None:
            self.type = type
        if size is not None:
            self.size = size



# Copyright 2014 Baidu, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file
# except in compliance with the License. You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the
# License is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND,
# either express or implied. See the License for the specific language governing permissions
# and limitations under the License.

"""
This module provide http request function for bce services.
"""
from future.utils import iteritems, iterkeys, itervalues
from builtins import str, bytes
import logging
import http.client
import sys
import time
import traceback

import baidubce
from baidubce import compat
from baidubce import utils
from baidubce.bce_response import BceResponse
from baidubce.exception import BceHttpClientError
from baidubce.exception import BceClientError
from baidubce.http import http_headers

_logger = logging.getLogger(__name__)


def _get_connection(protocol, host, port, connection_timeout_in_millis):
    """
    :param protocol
    :type protocol: baidubce.protocol.Protocol
    :param endpoint
    :type endpoint: str
    :param connection_timeout_in_millis
    :type connection_timeout_in_millis int
    """
    host = compat.convert_to_string(host)
    if protocol.name == baidubce.protocol.HTTP.name:
        return http.client.HTTPConnection(
            host=host, port=port, timeout=connection_timeout_in_millis / 1000)
    elif protocol.name == baidubce.protocol.HTTPS.name:
        return http.client.HTTPSConnection(
            host=host, port=port, timeout=connection_timeout_in_millis / 1000)
    else:
        raise ValueError(
            'Invalid protocol: %s, either HTTP or HTTPS is expected.' % protocol)


def _send_http_request(conn, http_method, uri, headers, body, send_buf_size):
    # putrequest() need that http_method and uri is Ascii on Py2 and unicode \
    # on Py3
    http_method = compat.convert_to_string(http_method)
    uri = compat.convert_to_string(uri)
    conn.putrequest(http_method, uri, skip_host=True, skip_accept_encoding=True)

    for k, v in iteritems(headers):
        k = utils.convert_to_standard_string(k)
        v = utils.convert_to_standard_string(v)
        conn.putheader(k, v)
    conn.endheaders()

    if body:
        if isinstance(body, (bytes, str)):
            conn.send(body)
        else:
            total = int(headers[http_headers.CONTENT_LENGTH])
            sent = 0
            while sent < total:
                size = total - sent
                if size > send_buf_size:
                    size = send_buf_size
                buf = body.read(size)
                if not buf:
                    raise BceClientError(
                        'Insufficient data, only %d bytes available while %s is %d' % (
                            sent, http_headers.CONTENT_LENGTH, total))
                conn.send(buf)
                sent += len(buf)

    return conn.getresponse()


def check_headers(headers):
    """
    check value in headers, if \n in value, raise
    :param headers:
    :return:
    """
    for k, v in iteritems(headers):
        if isinstance(v, (bytes, str)) and \
        b'\n' in compat.convert_to_bytes(v):
            raise BceClientError(r'There should not be any "\n" in header[%s]:%s' % (k, v))


def send_request(
        config,
        sign_function,
        response_handler_functions,
        http_method, path, body, headers, params, use_backup_endpoint=False):
    """
    Send request to BCE services.

    :param config
    :type config: baidubce.BceClientConfiguration

    :param sign_function:

    :param response_handler_functions:
    :type response_handler_functions: list

    :param request:
    :type request: baidubce.internal.InternalRequest

    :return:
    :rtype: baidubce.BceResponse
    """
    _logger.debug(b'%s request start: %s %s, %s, %s',
                  http_method, path, headers, params, body)
    headers = headers or {}

    user_agent = 'bce-sdk-python/%s/%s/%s' % (
        compat.convert_to_string(baidubce.SDK_VERSION), sys.version, sys.platform)
    user_agent = user_agent.replace('\n', '')
    user_agent = compat.convert_to_bytes(user_agent)
    headers[http_headers.USER_AGENT] = user_agent
 
    should_get_new_date = False
    if http_headers.BCE_DATE not in headers:
        should_get_new_date = True

    request_endpoint = config.endpoint
    if use_backup_endpoint:
        request_endpoint = config.backup_endpoint

    headers[http_headers.HOST] = request_endpoint

    if isinstance(body, str):
        body = body.encode(baidubce.DEFAULT_ENCODING)
    if not body:
        headers[http_headers.CONTENT_LENGTH] = 0
    elif isinstance(body, bytes):
        headers[http_headers.CONTENT_LENGTH] = len(body)
    elif http_headers.CONTENT_LENGTH not in headers:
        raise ValueError(b'No %s is specified.' % http_headers.CONTENT_LENGTH)
    # store the offset of fp body
    offset = None
    if hasattr(body, "tell") and hasattr(body, "seek"):
        offset = body.tell()

    protocol, host, port = utils.parse_host_port(request_endpoint, config.protocol)

    headers[http_headers.HOST] = host
    if port != config.protocol.default_port:
        headers[http_headers.HOST] += b':' + compat.convert_to_bytes(port)

    headers[http_headers.AUTHORIZATION] = sign_function(
        config.credentials, http_method, path, headers, params)

    encoded_params = utils.get_canonical_querystring(params, False)
    if len(encoded_params) > 0:
        uri = path + b'?' + encoded_params
    else:
        uri = path
    check_headers(headers)

    retries_attempted = 0
    errors = []
    while True:
        conn = None
        try:
            # restore the offset of fp body when retrying
            if should_get_new_date is True:
                headers[http_headers.BCE_DATE] = utils.get_canonical_time()

            headers[http_headers.AUTHORIZATION] = sign_function(
                config.credentials, http_method, path, headers, params)

            if retries_attempted > 0 and offset is not None:
                body.seek(offset)
            
            conn = _get_connection(protocol, host, port, config.connection_timeout_in_mills)

            _logger.debug('request args:method=%s, uri=%s, headers=%s,patams=%s, body=%s',
                    http_method, uri, headers, params, body)

            http_response = _send_http_request(
                conn, http_method, uri, headers, body, config.send_buf_size)
            
            headers_list = http_response.getheaders()

            # on py3 ,values of headers_list is decoded with ios-8859-1 from
            # utf-8 binary bytes

            # headers_list[*][0] is lowercase on py2
            # headers_list[*][0] is raw value py3
            if compat.PY3 and isinstance(headers_list, list):
                temp_heads = []
                for k, v in headers_list:
                    k = k.encode('latin-1').decode('utf-8')
                    v = v.encode('latin-1').decode('utf-8')
                    k = k.lower()
                    temp_heads.append((k, v))
                headers_list = temp_heads

            _logger.debug(
                'request return: status=%d, headers=%s' % (http_response.status, headers_list))
            response = BceResponse()
            response.set_metadata_from_headers(dict(headers_list))

            for handler_function in response_handler_functions:
                if handler_function(http_response, response):
                    break
            return response
        except Exception as e:
            if conn is not None:
                conn.close()

            # insert ">>>>" before all trace back lines and then save it
            errors.append('\n'.join('>>>>' + line for line in traceback.format_exc().splitlines()))

            if config.retry_policy.should_retry(e, retries_attempted):
                delay_in_millis = config.retry_policy.get_delay_before_next_retry_in_millis(
                    e, retries_attempted)
                time.sleep(delay_in_millis / 1000.0)
            else:
                raise BceHttpClientError('Unable to execute HTTP request. Retried %d times. '
                                         'All trace backs:\n%s' % (retries_attempted,
                                                                   '\n'.join(errors)), e)

        retries_attempted += 1



# Copyright (c) 2014 Baidu.com, Inc. All Rights Reserved
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file
# except in compliance with the License. You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the
# License is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND,
# either express or implied. See the License for the specific language governing permissions
# and limitations under the License.

"""
This module provides a client class for CDN.
"""

import copy
import logging
import uuid

from baidubce import bce_base_client
from baidubce.auth import bce_v1_signer
from baidubce.http import bce_http_client
from baidubce.http import handler
from baidubce.http import http_methods
from baidubce.utils import required

_logger = logging.getLogger(__name__)


class DumapClient(bce_base_client.BceBaseClient):
    """
    DumapClient
    """

    def __init__(self, config=None):
        bce_base_client.BceBaseClient.__init__(self, config)

    @required(app_id=(bytes, str), uri=(bytes, str), params=dict)
    def call_open_api(self, app_id=None, uri=None, params=None, body=None, method=b'GET', config=None):
        """
        call open_api
        :param app_id: app_id
        :type app_id: string
        :param uri: open api uri
        :type uri: string
        :param params: dict
        :type params:request params
        :param body: request body (default: None)
        :type body: string
        :param method: http method (default GET)
        :type method: http_methods
        :param config: None
        :type config: baidubce.BceClientConfiguration

        :return:
        :rtype: baidubce.bce_response.BceResponse
                """
        response = self._send_request(
            http_method=method,
            path=uri,
            params=params,
            body=body,
            headers={b'x-app-id': app_id},
            config=config
        )

        if response.body:
            return response.body.decode("utf-8")
        else:
            return response

    @staticmethod
    def _merge_config(self, config):
        if config is None:
            return self.config
        else:
            new_config = copy.copy(self.config)
            new_config.merge_non_none_values(config)
            return new_config

    def _send_request(
            self, http_method, path,
            body=None, headers=None, params=None,
            config=None):
        config = self._merge_config(self, config)
        headers[b'x-bce-request-id'] = uuid.uuid4()
        headers[b'Content-Type'] = b'application/json;charset=utf-8'

        return bce_http_client.send_request(
            config, sign_wrapper([b'host', b'x-bce-date', b'x-bce-request-id', b'x-app-id']),
            [handler.parse_error, parse_none],
            http_method, path, body, headers, params)


def sign_wrapper(headers_to_sign):
    """wrapper the bce_v1_signer.sign()."""

    def _wrapper(credentials, http_method, path, headers, params):
        return bce_v1_signer.sign(credentials, http_method, path, headers, params,
                                  headers_to_sign=headers_to_sign)

    return _wrapper


def parse_none(http_response, response):
    """If the body is not empty, convert it to a python object and set as the value of
    response.body. http_response is always closed if no error occurs.

    :param http_response: the http_response object returned by HTTPConnection.getresponse()
    :type http_response: httplib.HTTPResponse

    :param response: general response object which will be returned to the caller
    :type response: baidubce.BceResponse

    :return: always true
    :rtype bool
    """
    body = http_response.read()
    if body:
        response.body = body
    http_response.close()
    return True



# Copyright (c) 2014 Baidu.com, Inc. All Rights Reserved
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file
# except in compliance with the License. You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the
# License is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND,
# either express or implied. See the License for the specific language governing permissions
# and limitations under the License.

"""
This module provides a client class for BBC.
"""
import copy
import json
import logging
import random
import string
import uuid

from baidubce import bce_base_client
from baidubce.auth import bce_v1_signer
from baidubce.http import bce_http_client
from baidubce.http import handler
from baidubce.http import http_methods
from baidubce.utils import aes128_encrypt_16char_key
from baidubce.utils import required
from baidubce.services.bbc import bbc_model
from baidubce import compat

_logger = logging.getLogger(__name__)

default_billing_to_purchase_created = bbc_model.Billing('Postpaid')



class BbcClient(bce_base_client.BceBaseClient):
    """
    Bbc client sdk
    """

    def __init__(self, config=None):
        bce_base_client.BceBaseClient.__init__(self, config)

    def _merge_config(self, config=None):
        if config is None:
            return self.config
        else:
            new_config = copy.copy(self.config)
            new_config.merge_non_none_values(config)
            return new_config

    def _send_request(self, http_method, path,
                      body=None, headers=None, params=None,
                      config=None, body_parser=None, api_version=b'/v1'):
        config = self._merge_config(config)
        if body_parser is None:
            body_parser = handler.parse_json

        return bce_http_client.send_request(
            config, bce_v1_signer.sign, [handler.parse_error, body_parser],
            http_method, api_version + path, body, headers, params)



    @required(flavor_id = (bytes, str), image_id = (bytes, str), raid_id = (bytes, str))
    def create_instance(self, flavor_id, image_id, raid_id, root_disk_size_in_gb=20, purchase_count=1,
                        zone_name=None, subnet_id=None, billing=None, name=None, admin_pass=None,
                        auto_renew_time_unit=None, auto_renew_time=0, deploy_set_id=None, enable_ht=False,
                        security_group_id=None, client_token=None, config=None):

        """
        Create a bbc instance with the specified options.
        You must fill the field of clientToken, which is especially for keeping idempotent.
        This is an asynchronous interface.

        :param flavor_id:
            The id of flavor, list all available flavors in BbcClient.list_flavors.
        :type flavor_id: string

        :param image_id:
            The id of image, list all available images in BbcClient.list_images.
        :type image_id: string

        :param raid_id:
            The id of raid, list all available raids in BbcClient.get_flavor_raid.
        :type raid_id: string

        :param root_disk_size_in_gb:
            System disk size of the physical machine to be created
        :type root_disk_size_in_gb: int

        :param purchase_count:
            The number of instances to buy, the default value is 1.
        :type purchase_count: int

        :param zone_name:
            The optional parameter to specify the available zone for the instance.
            See more detail through list_zones method
        :type zone_name: string

        :param subnet_id:
            The optional parameter to specify the id of subnet from vpc, optional param
             default value is default subnet from default vpc
        :type subnet_id: string

        :param billing:
            Billing information.
        :type billing: bbc_model.Billing

        :param name:
            The optional parameter to desc the instance that will be created.
        :type name: string

        :param enable_ht:
            The optional parameter to enable instance hyperthread.
        :type name: bool

        :param admin_pass:
            The optional parameter to specify the password for the instance.
            If specify the adminPass,the adminPass must be a 8-16 characters String
            which must contains letters, numbers and symbols.
            The symbols only contains "!@#$%^*()".
            The adminPass will be encrypted in AES-128 algorithm
            with the substring of the former 16 characters of user SecretKey.
            If not specify the adminPass, it will be specified by an random string.
            See more detail on
            https://cloud.baidu.com/doc/BBC/s/3jwvxu9iz#%E5%AF%86%E7%A0%81%E5%8A%A0%E5%AF%86%E4%BC%A0%E8%BE%93%E8%A7%84%E8%8C%83
        :type admin_pass: string

        :param auto_renew_time_unit
            The parameter to specify the unit of the auto renew time.
            The auto renew time unit can be "month" or "year".
            The default value is "month".
        :type auto_renew_time_unit: string

        :param auto_renew_time
            The parameter to specify the auto renew time, the default value is 0.
        :type auto_renew_time: string

        :deploy_set_id:
            The id of the deploy set
        :type deploy_set_id: string

        :param security_group_id:
            The optional parameter to specify the securityGroupId of the instance
            vpcId of the securityGroupId must be the same as the vpcId of subnetId
        :type security_group_id: string

        :param client_token:
            An ASCII string whose length is less than 64.
            The request will be idempotent if client token is provided.
            If the clientToken is not specified by the user,
            a random String generated by default algorithm will be used.
            See more detail at
            https://bce.baidu.com/doc/BCC/API.html#.E5.B9.82.E7.AD.89.E6.80.A7
        :type client_token: string

        :return:
        :rtype baidubce.bce_response.BceResponse
        """
        path = b'/instance'
        params = {}
        if client_token is None:
            params['clientToken'] = generate_client_token()
        else:
            params['clientToken'] = client_token


        if root_disk_size_in_gb == 0:
            root_disk_size_in_gb = 20
        if purchase_count < 1:
            purchase_count = 1

        if billing is None:
            billing = default_billing_to_purchase_created

        body = {
            'flavorId': flavor_id,
            'imageId': image_id,
            'raidId': raid_id,
            'rootDiskSizeInGb': root_disk_size_in_gb,
            'purchaseCount': purchase_count,
            'billing': billing.__dict__
        }
        if zone_name is not None:
            body['zoneName'] = zone_name
        if subnet_id is not None:
            body['subnetId'] = subnet_id
        if security_group_id is not None:
            body['securityGroupId'] = security_group_id
        if name is not None:
            body['name'] = name
        if deploy_set_id is not None:
            body['deploySetId'] = deploy_set_id
        if enable_ht:
            body['enableHt'] = True
        if admin_pass is not None:
            secret_access_key = self.config.credentials.secret_access_key
            cipher_admin_pass = aes128_encrypt_16char_key(admin_pass, secret_access_key)
            body['adminPass'] = cipher_admin_pass
        if auto_renew_time_unit is None:
            body['autoRenewTimeUnit'] = "month"
        else:
            body['autoRenewTimeUnit'] = auto_renew_time_unit
        if auto_renew_time != 0:
            body['autoRenewTime'] = auto_renew_time
        return self._send_request(http_methods.POST, path, json.dumps(body),
                                  params=params, config=config)


    def list_instances(self, marker=None, max_keys=1000, internal_ip=None, config=None):
        """
        Return a list of instances owned by the authenticated user.

        :param marker:
            The optional parameter marker specified in the original request to specify
            where in the results to begin listing.
            Together with the marker, specify the list result which listing should begin.
            If the marker is not specified, the list result will listing from the first one.
        :type marker: string

        :param max_keys:
            The optional parameter to specify the max number of list result to return.
            The default value is 1000.
        :type max_keys: int

        :param internal_ip:
            The identified internal ip of instance.
        :type internal_ip: string

        :return:
        :rtype baidubce.bce_response.BceResponse
        """

        path = b'/instance'
        params = {}

        if marker is not None:
            params['marker'] = marker
        if max_keys is not None:
            params['maxKeys'] = max_keys
        if internal_ip is not None:
            params['internalIp'] = internal_ip

        return self._send_request(http_methods.GET, path, params=params, config=config)



    @required(instance_id=(bytes, str))
    def get_instance(self, instance_id, contains_failed=False, config=None):
        """
        Get the detailed information of specified instance.

        :param instance_id:
            The id of instance.
        :type instance_id: string

        :param contains_failed:
            The optional parameters to get the failed message.If true, it means get the failed message.
        :type contains_failed: boolean

        :return:
        :rtype baidubce.bce_response.BceResponse
        """
        instance_id = compat.convert_to_bytes(instance_id)
        path = b'/instance/%s' % instance_id
        params = {}
        if contains_failed:
            params['containsFailed'] = contains_failed
        return self._send_request(http_methods.GET, path, params=params, config=config)

    @required(instance_id=(bytes, str))
    def start_instance(self, instance_id, config=None):
        """
        Starting the instance owned by the user.
        You can start the instance only when the instance is Stopped,
        otherwise, it's will get 409 errorCode.
        This is an asynchronous interface.

        :param instance_id: id of instance proposed to start
        :type instance_id: string

        :return:
        :rtype baidubce.bce_response.BceResponse
        """
        instance_id = compat.convert_to_bytes(instance_id)
        path = b'/instance/%s' % instance_id
        params = {
            'start': None
        }
        return self._send_request(http_methods.PUT, path, params=params, config=config)

    @required(instance_id=(bytes, str))
    def stop_instance(self, instance_id, force_stop=False, config=None):
        """
        Stopping the instance owned by the user.
        You can stop the instance only when the instance is Running,
        otherwise, it's will get 409 errorCode.
        This is an asynchronous interface.

        :param instance_id:
            The id of instance.
        :type instance_id: string

        :param force_stop:
            The optional parameter to stop the instance forcibly.If true,
            it will stop the instance just like power off immediately
            and it may result in losing important data which have not been written to disk.
        :type force_stop: boolean

        :return:
        :rtype baidubce.bce_response.BceResponse
        """
        instance_id = compat.convert_to_bytes(instance_id)
        path = b'/instance/%s' % instance_id
        body = {
            'forceStop': force_stop
        }
        params = {
            'stop': None
        }
        return self._send_request(http_methods.PUT, path, json.dumps(body),
                                  params=params, config=config)

    @required(instance_id=(bytes, str))
    def reboot_instance(self, instance_id, force_stop=False, config=None):
        """
        Rebooting the instance owned by the user.
        You can reboot the instance only when the instance is Running,
        otherwise, it's will get 409 errorCode.
        This is an asynchronous interface.

        :param instance_id:
            The id of instance.
        :type instance_id: string

        :param force_stop:
            The optional parameter to stop the instance forcibly.If true,
            it will stop the instance just like power off immediately
            and it may result in losing important data which have not been written to disk.
        :type force_stop: boolean

        :return:
        :rtype baidubce.bce_response.BceResponse
        """
        instance_id = compat.convert_to_bytes(instance_id)
        path = b'/instance/%s' % instance_id
        body = {
            'forceStop': force_stop
        }
        params = {
            'reboot': None
        }
        return self._send_request(http_methods.PUT, path, json.dumps(body),
                                  params=params, config=config)

    @required(instance_id=str)
    def batch_add_ip(self, instance_id, private_ips=None, secondary_private_ip_address_count=None, config=None):
        """
        batch_add_ip
        :param instance_id:
        :param private_ips:
        :param secondary_private_ip_address_count:
        :param config:
        :return:
        :rtype baidubce.bce_response.BceResponse
        """
        path = b'/instance/batchAddIp'
        body = {
            'instanceId': instance_id,
        }
        if private_ips is not None:
            body['privateIps'] = private_ips
        if secondary_private_ip_address_count is not None:
            body['secondaryPrivateIpAddressCount'] = secondary_private_ip_address_count
        params = {

        }
        return self._send_request(http_methods.PUT, path, json.dumps(body),
                                  params=params, config=config)

    @required(instance_id=str, private_ips=list)
    def batch_delete_ip(self, instance_id, private_ips, config=None):
        """
        :param instance_id:
        :param private_ips:
        :param config:
        :return:
        """
        path = b'/instance/batchDelIp'
        body = {
            'instanceId': instance_id,
            'privateIps': private_ips,
        }
        params = {

        }
        return self._send_request(http_methods.PUT, path, json.dumps(body),
                                  params=params, config=config)

    @required(instance_id=(bytes, str),
              renew_time_unit=(bytes, str),
              renew_time=int)
    def create_auto_renew_rules(self, instance_id, renew_time_unit, renew_time, config=None):
        """
        Creating auto renew rules for the bbc.
        It only works for the prepaid bbc.

        :param instance_id:
            The id of instance.
        :type instance_id: string

        :param renew_time_unit
            The parameter to specify the unit of the renew time.
            The renew time unit can be "month" or "year".
        :type renew_time_unit: string

        :param renew_time
            The parameter to specify the renew time.
        :type renew_time: int
        """
        instance_id = compat.convert_to_bytes(instance_id)
        path = b'/instance/batchCreateAutoRenewRules'
        body = {
            'instanceId': instance_id,
            'renewTimeUnit': renew_time_unit,
            'renewTime': renew_time
        }
        params = {

        }
        return self._send_request(http_methods.POST, path, json.dumps(body),
                                  params=params, config=config)

    @required(instance_id=(bytes, str))
    def delete_auto_renew_rules(self, instance_id, config=None):
        """
        Deleting auto renew rules for the bbc.
        It only works for the prepaid bbc.

        :param instance_id:
            The id of instance.
        :type instance_id: string
        """
        instance_id = compat.convert_to_bytes(instance_id)
        path = b'/instance/batchDeleteAutoRenewRules'
        body = {
            'instanceId': instance_id
        }
        params = {

        }
        return self._send_request(http_methods.POST, path, json.dumps(body),
                                  params=params, config=config)

    @required(instance_id=(bytes, str),
              name=(bytes, str))
    def modify_instance_name(self, instance_id, name, config=None):
        """
        Modifying the name of the instance.
        You can modify the instance name only when the instance is Running or Stopped ,
        otherwise, it's will get 409 errorCode.

        :param instance_id:
            The id of instance.
        :type instance_id: string

        :param name:
            The new value for instance's name.
        :type name: string

        :return:
        :rtype baidubce.bce_response.BceResponse
        """
        instance_id = compat.convert_to_bytes(instance_id)
        path = b'/instance/%s' % instance_id
        body = {
            'name': name
        }
        params = {
            'rename': None
        }
        return self._send_request(http_methods.PUT, path, json.dumps(body),
                                  params=params, config=config)



    def modify_instance_desc(self, instance_id, desc, config=None):
        """
        Modifying the description  of the instance.
        You can modify the description only when the instance is Running or Stopped ,
        otherwise, it's will get 409 errorCode.

        :param instance_id:
            The id of instance.
        :type instance_id: string

        :param desc:
            The new description of the instance.
        :type name: string

        :return:
        :rtype baidubce.bce_response.BceResponse
        """
        instance_id = compat.convert_to_bytes(instance_id)
        path = b'/instance/%s' % instance_id
        body = {
            'desc': desc
        }
        params = {
            'updateDesc': None
        }
        return self._send_request(http_methods.PUT, path, json.dumps(body),
                                  params=params, config=config)

    @required(instance_id=(bytes, str),
              image_id=(bytes, str),
              admin_pass=(bytes, str))
    def rebuild_instance(self, instance_id, image_id, admin_pass,
                         is_preserve_data=True, raid_id=None, sys_root_size=20, config=None):
        """
        Rebuilding the instance owned by the user.
        After rebuilding the instance,
        all of snapshots created from original instance system disk will be deleted,
        all of customized images will be saved for using in the future.
        This is an asynchronous interface

        :param instance_id:
            The id of instance.
        :type instance_id: string

        :param image_id:
            The id of the image which is used to rebuild the instance.
        :type image_id: string

        :param admin_pass:
            The admin password to login the instance.
            The admin password will be encrypted in AES-128 algorithm
            with the substring of the former 16 characters of user SecretKey.
            See more detail on
            https://cloud.baidu.com/doc/BBC/s/3jwvxu9iz#%E5%AF%86%E7%A0%81%E5%8A%A0%E5%AF%86%E4%BC%A0%E8%BE%93%E8%A7%84%E8%8C%83
        :type admin_pass: string

        :param is_preserve_data:
            Whether or not to retain data, the default is true. The raid_id and
            sys_root_size fields do not take effect when the value is true
        :type is_preserve_data: bool

        :param raid_id:
            The id of raid. See more details on
            https://cloud.baidu.com/doc/BBC/s/Bjwvxu9ul#%E6%9F%A5%E8%AF%A2raid
        :type raid_id: string

        :param sys_root_size:
            System root partition size, default is 20G, value range is 20-100
        :type sys_root_size: int

        :return:
        :rtype baidubce.bce_response.BceResponse
        """
        secret_access_key = self.config.credentials.secret_access_key
        cipher_admin_pass = aes128_encrypt_16char_key(admin_pass, secret_access_key)
        instance_id = compat.convert_to_bytes(instance_id)
        path = b'/instance/%s' % instance_id
        body = {
            'imageId': image_id,
            'adminPass': cipher_admin_pass,
            'isPreserveData': is_preserve_data,
            'raidId': raid_id,
            'sysRootSize': sys_root_size
        }
        params = {
            'rebuild': None
        }
        return self._send_request(http_methods.PUT, path, json.dumps(body),
                                  params=params, config=config)

    @required(instance_id=(bytes, str))
    def release_instance(self, instance_id, config=None):
        """
        Releasing the instance owned by the user.
        Only the Postpaid instance or Prepaid which is expired can be released.
        After releasing the instance,
        all of the data will be deleted.
        all of snapshots created from original instance system disk will be deleted,
        all of customized images created from original instance system disk will be reserved.

        :param instance_id:
            The id of instance.
        :type instance_id: string

        :return:
        :rtype baidubce.bce_response.BceResponse
        """
        instance_id = compat.convert_to_bytes(instance_id)
        api_version = b'/v2'
        path = b'/instance/%s' % instance_id
        return self._send_request(http_methods.DELETE, path, config=config, api_version=api_version)

    @required(instance_id=(bytes, str),
              admin_pass=(bytes, str))
    def modify_instance_password(self, instance_id, admin_pass, config=None):
        """
        Modifying the password of the instance.
        You can change the instance password only when the instance is Running or Stopped ,
        otherwise, it's will get 409 errorCode.
        This is an asynchronous interface.

        :param instance_id:
            The id of instance.
        :type instance_id: string

        :param admin_pass:
            The new password to update.
            The adminPass will be encrypted in AES-128 algorithm
            with the substring of the former 16 characters of user SecretKey.
        :type admin_pass: string

        :return:
        :rtype baidubce.bce_response.BceResponse
        """
        secret_access_key = self.config.credentials.secret_access_key
        cipher_admin_pass = aes128_encrypt_16char_key(admin_pass, secret_access_key)
        instance_id = compat.convert_to_bytes(instance_id)
        path = b'/instance/%s' % instance_id
        body = {
            'adminPass': cipher_admin_pass
        }
        params = {
            'changePass': None
        }
        return self._send_request(http_methods.PUT, path, json.dumps(body),
                                  params=params, config=config)

    @required(bbc_ids = list)
    def get_vpc_subnet(self, bbc_ids, config=None):
        """
        Query VPC / Subnet information by BBC instance id

        :param bbc_ids:
            List of BBC instance IDs that need to query VPC / Subnet information
        :type bbc_ids: list

        :return:
        :rtype baidubce.bce_response.BceResponse
        """
        path = b'/vpcSubnet'
        body = {
            'bbcIds': bbc_ids
        }

        return self._send_request(http_methods.POST, path, body=json.dumps(body), config = config)

    @required(instance_id=(bytes, str), change_tags = list)
    def unbind_tags(self, instance_id, change_tags, config=None):
        """
        Unbind the tags of existing instances

        :param instance_id:
            The id of instance.
        :type instance_id: string

        :param change_tags:
            List of tags to be unbind
        :type change_tags: list

        :return:
        :rtype baidubce.bce_response.BceResponse

        """
        instance_id = compat.convert_to_bytes(instance_id)
        path = b'/instance/%s/tag' % instance_id

        params = {
            'unbind': None
        }

        body = {
            'changeTags': change_tags
        }

        return self._send_request(http_methods.PUT, path, body=json.dumps(body),
                                  params=params, config=config)


    def list_flavors(self, config=None):
        """
        :return:
        :rtype baidubce.bce_response.BceResponse

        """
        path = b'/flavor'
        return self._send_request(http_methods.GET, path, config = config)

    @required(flavor_id = (bytes, str))
    def get_flavor(self, flavor_id, config=None):
        """
        :param flavor_id:
            The id of flavor.
        :type flavor_id: string

        :return:
        :rtype baidubce.bce_response.BceResponse

        """
        flavor_id = compat.convert_to_bytes(flavor_id)
        path = b'/flavor/%s' % flavor_id
        return self._send_request(http_methods.GET, path, config=config)

    @required(flavor_id=(bytes, str))
    def get_flavor_raid(self, flavor_id, config=None):
        """
        :param flavor_id:
            The id of flavor.
        :type flavor_id: string

        :return:
        :rtype baidubce.bce_response.BceResponse
        """
        flavor_id = compat.convert_to_bytes(flavor_id)
        path = b'/flavorRaid/%s' % flavor_id
        return self._send_request(http_methods.GET, path, config=config)

    @required(image_name=(bytes, str),
              instance_id=(bytes, str))
    def create_image_from_instance_id(self,
                                      image_name,
                                      instance_id,
                                      client_token=None,
                                      config=None):
        """
        Creating a customized image which can be used for creating instance.
        You can create an image from an instance with this method.
        While creating an image from an instance, the instance must be Running or Stopped,
        otherwise, it's will get 409 errorCode.
        This is an asynchronous interface.

        :param image_name:
            The name for the image that will be created.
            The name length from 1 to 65,only contains letters,digital and underline.
        :type image_name: string

        :param instance_id:
            The optional parameter specify the id of the instance which will be used to create the new image.
            When instanceId and snapshotId are specified ,only instanceId will be used.
        :type instance_id: string

        :param client_token:
            An ASCII string whose length is less than 64.
            The request will be idempotent if client token is provided.
            If the clientToken is not specified by the user,
            a random String generated by default algorithm will be used.
            See more detail at
            https://bce.baidu.com/doc/BCC/API.html#.E5.B9.82.E7.AD.89.E6.80.A7
        :type client_token: string

        :return:
        :rtype baidubce.bce_response.BceResponse
        """
        path = b'/image'
        params = None
        if client_token is None:
            params = {
                'clientToken': generate_client_token()
            }
        else:
            params = {
                'clientToken': client_token
            }
        body = {
            'imageName': image_name,
            'instanceId': instance_id
        }

        return self._send_request(http_methods.POST, path, json.dumps(body),
                                  params=params, config=config)

    def list_images(self, image_type='All', marker=None, max_keys=1000, config=None):
        """
        Listing images owned by the authenticated user.

        :param image_type:
            The optional parameter to filter image to list.
            See more detail at
            https://bce.baidu.com/doc/BCC/API.html#ImageType"
        :type image_type: menu{'All', System', 'Custom', 'Integration'}

        :param marker:
            The optional parameter marker specified in the original request to specify
            where in the results to begin listing.
            Together with the marker, specifies the list result which listing should begin.
            If the marker is not specified, the list result will listing from the first one.
        :type marker: string

        :param max_keys:
            The optional parameter to specify the max number of list result to return.
            The default value is 1000.
        :type max_keys: int

        :return:
        :rtype baidubce.bce_response.BceResponse
        """
        path = b'/image'
        params = {
            'imageType': image_type
        }
        if marker is not None:
            params['marker'] = marker
        if max_keys is not None:
            params['maxKeys'] = max_keys

        return self._send_request(http_methods.GET, path, params=params, config=config)

    @required(image_id=(bytes, str))
    def get_image(self, image_id, config=None):
        """
        Get the detail information of specified image.

        :param image_id:
            The id of image.
        :type image_id: string

        :return:
        :rtype baidubce.bce_response.BceResponse
        """
        image_id = compat.convert_to_bytes(image_id)
        path = b'/image/%s' % image_id
        return self._send_request(http_methods.GET, path, config=config)

    @required(image_id=(bytes, str))
    def delete_image(self, image_id, config=None):
        """
        Deleting the specified image.
        Only the customized image can be deleted,
        otherwise, it's will get 403 errorCode.

        :param image_id:
            The id of image.
        :type image_id: string

        :return:
        :rtype baidubce.bce_response.BceResponse
        """
        image_id = compat.convert_to_bytes(image_id)
        path = b'/image/%s' % image_id
        return self._send_request(http_methods.DELETE, path, config=config)

    def get_operation_log(self, marker=None, max_keys=100, start_time=None, end_time=None, config=None):
        """
        Querying information about physical machine operation logs

        :param marker:
            The optional parameter marker specified in the original request to specify
            where in the results to begin listing.
            Together with the marker, specifies the list result which listing should begin.
            If the marker is not specified, the list result will listing from the first one.
        :type marker: string

        :param max_keys:
            The optional parameter to specify the max number of list result to return.
            The default value is 1000.
        :type max_keys: int

        :param start_time:
            The start time of the physical machine operation (UTC time),
            the format is yyyy-MM-dd'T'HH: mm: ss'Z ', if it is empty, query the operation log of the day
        :type start_time: string

        :params end_time
            The end time of the physical machine operation (UTC time),
            the format is yyyy-MM-dd'T'HH: mm: ss'Z ', if it is empty, query the operation log of the day
        :type end_time: string

        :return:
        :rtype baidubce.bce_response.BceResponse
        """
        path = b'/operationLog'
        params = {}
        if marker is not None:
            params['marker'] = marker
        if max_keys is not None:
            params['maxKeys'] = max_keys
        if start_time is not None:
            params['startTime'] = start_time
        if end_time is not None:
            params['endTime'] = end_time

        return self._send_request(http_methods.GET, path, params=params, config=config)

    @required(concurrency=int, strategy=(bytes, str))
    def create_deploy_set(self, concurrency, strategy, name=None, desc=None, client_token=None, config=None):
        """
        Create a deploy set based on a specified policy and concurrency

        :param concurrency:
            Deployment set concurrency, range [1,5]
        :type concurrency: int

        :param strategy:
            Deployment set strategy, currently BBC strategy only supports: "tor_ha", "host_ha"
        :type strategy: string

        :param name:
            Deployment set name, supports uppercase and lowercase letters, numbers,
            Chinese, and -_ /. Special characters, must start with a letter, length 1-65
        :type name: string

        :param desc:
            Deployment set description
        :type desc: string

        :param client_token:
            An ASCII string whose length is less than 64.
            The request will be idempotent if client token is provided.
            If the clientToken is not specified by the user,
            a random String generated by default algorithm will be used.
            See more detail at
            https://bce.baidu.com/doc/BCC/API.html#.E5.B9.82.E7.AD.89.E6.80.A7
        :type client_token: string

        :return:
        :rtype baidubce.bce_response.BceResponse
        """
        path = b'/deployset'
        params = {}
        if client_token is None:
            params['clientToken'] = generate_client_token()
        else:
            params['clientToken'] = client_token
        body = {
            "strategy": strategy,
            "concurrency": concurrency
        }
        if name is not None:
            body["name"] = name
        if desc is not None:
            body['desc'] = desc
        return self._send_request(http_methods.POST, path, json.dumps(body),
                                  params=params, config=config)

    def list_deploy_sets(self, config=None):
        """
        List all deploy sets

        :return:
        :rtype baidubce.bce_response.BceResponse
        """
        path = b'/deployset'
        return self._send_request(http_methods.GET, path, config=config)

    @required(deploy_set_ids=(bytes, str))
    def get_deploy_set(self, deploy_set_id, config=None):
        """
        Get the specified deploy set

        :param deploy_set_id:
            The id of the deploy set
        :type deploy_set_id: String

        :return:
        :rtype baidubce.bce_response.BceResponse
        """
        deploy_set_id = compat.convert_to_bytes(deploy_set_id)
        path = b'/deployset/%s' % deploy_set_id
        return self._send_request(http_methods.GET, path, config=config)

    @required(deploy_set_ids=(bytes, str))
    def delete_deploy_set(self, deploy_set_id, config=None):
        """
        Delete the specified deploy sets

        :param deploy_set_ids:
            The ids of the deploy sets you want to delete
        :type deploy_set_ids: list

        :return:
        :rtype baidubce.bce_response.BceResponse
        """
        deploy_set_id = compat.convert_to_bytes(deploy_set_id)
        path = b'/deployset/%s' % deploy_set_id
        return self._send_request(http_methods.DELETE, path, config=config)









def generate_client_token_by_uuid():
    """
    The default method to generate the random string for client_token
    if the optional parameter client_token is not specified by the user.
    :return:
    :rtype string
    """
    return str(uuid.uuid4())


def generate_client_token_by_random():
    """
    The alternative method to generate the random string for client_token
    if the optional parameter client_token is not specified by the user.
    :return:
    :rtype string
    """
    client_token = ''.join(random.sample(string.ascii_letters + string.digits, 36))
    return client_token


generate_client_token = generate_client_token_by_uuid






# Copyright (c) 2014 Baidu.com, Inc. All Rights Reserved
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file
# except in compliance with the License. You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the
# License is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND,
# either express or implied. See the License for the specific language governing permissions
# and limitations under the License.

"""
This module provides a client for STS.
"""

from future.utils import iteritems
import copy
import http.client
import os
import json
import logging
import shutil
from builtins import str
from builtins import bytes

import baidubce
from baidubce import bce_client_configuration
from baidubce import utils
from baidubce.auth import bce_v1_signer
from baidubce.bce_base_client import BceBaseClient
from baidubce.exception import BceClientError
from baidubce.exception import BceServerError
from baidubce.exception import BceHttpClientError
from baidubce.http import bce_http_client
from baidubce.http import handler
from baidubce.http import http_content_types
from baidubce.http import http_headers
from baidubce.http import http_methods
from baidubce.services import sts
from baidubce.utils import required


_logger = logging.getLogger(__name__)


class StsClient(BceBaseClient):
    """
    sdk client
    """
    def __init__(self, config=None):
        BceBaseClient.__init__(self, config)

    def get_session_token(self, acl, duration_seconds=None):    
        """
        :type duration_seconds: int
        :param duration_seconds: None
        :token effective period

        :type id: string
        :param id: None
        :acl id

        :type acl: dict
        :param acl: None
        :acl

        :return:
            **HttpResponse**
        """

        params = None

        if duration_seconds is not None:
            if isinstance(duration_seconds, int):
                params = {b'durationSeconds': duration_seconds}

        if acl is None:
            body = None
        else:
            if not isinstance(acl, dict):
                raise TypeError(b'acl should be dict')
            if 'id' in acl:
                body = json.dumps(acl)
            else:
                body = json.dumps(acl, default=self._dump_acl_object)

        return self._send_request(
            http_methods.POST,
            body=body,
            headers={http_headers.CONTENT_TYPE: http_content_types.JSON},
            params=params)

    @staticmethod
    def _dump_acl_object(acl):
        result = {}
        for k, v in iteritems(acl.__dict__):
            if not k.startswith('_'):
                result[k] = v
        return result

    def _merge_config(self, config):
        if config is None:
            return self.config
        else:
            new_config = copy.copy(self.config)
            new_config.merge_non_none_values(config)
            return new_config

    def _send_request(self, 
                      http_method,
                      body=None, 
                      headers=None, 
                      params=None,
                      config=None,
                      body_parser=None):
        config = self._merge_config(config)
        path = sts.URL_PREFIX + b"sessionToken"
        if body_parser is None:
            body_parser = handler.parse_json

        return bce_http_client.send_request(
            config, bce_v1_signer.sign, [handler.parse_error, body_parser],
            http_method, path, body, headers, params)



# Copyright 2014 Baidu, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file
# except in compliance with the License. You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the
# License is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND,
# either express or implied. See the License for the specific language governing permissions
# and limitations under the License.
"""
This module provides models for CFC-SDK.
"""

import json

DUEDGE_TRIGGER = 'duedge'

DUEROS_TRIGGER = 'dueros'

CRONTAB_TRIGGER = 'cfc-crontab-trigger/v1/'

HTTP_TRIGGER = 'cfc-http-trigger/v1/CFCAPI'

CDN_TRIGGER = 'cdn'

BOS_TRIGGER = 'bos'


class AbstractTriggerDataModel(object):
    """Base class for all trigger data."""

    def serialize(self):
        """
        serialize

        :return
        :rtype dict
        """
        d = vars(self)
        ret = {}
        for k in d:
            if d[k] is not None:
                ret[k] = d[k]
        return ret

    def get_trigger_source(self):
        """
        trigger source

        :return
        :rtype string
        """
        return ''

    def __repr__(self):
        return '%s' % json.dumps(self.serialize())


class CrontabTriggerData(AbstractTriggerDataModel):
    """
    Crontab Trigger Data
    :param brn  (required) The url path.
    :type brn string
    :param name  (required) The trigger name. 1-30 length.Pattern: ^[a-zA-Z0-9-_]+$
    :type name string.The name of the trigger that you are creating or updating
    :param schedule_expression  (required) Schedule expression.The details see
        https://cloud.baidu.com/doc/CFC/s/Zjxl9lbed.
        For example, "cron(0 * * * *)" or "rate(10 minutes)".
    :type schedule_expression string.
    :param enabled.
    :type enabled bool. Enables the trigger.
    :param custom_input.
    :type custom_input json.
    """

    def __init__(self, brn=None, name=None, schedule_expression=None, enabled=False, custom_input=None):
        self.Input = custom_input
        self.Brn = brn
        self.Name = name
        self.ScheduleExpression = schedule_expression
        self.UUID = None
        if enabled:
            self.Enabled = 'Enabled'
        else:
            self.Enabled = 'Disabled'

    def set_status(self, enabled=False):
        """
        set crontab status

        :param enabled.
        :type enabled bool. Enables the trigger.
        :return
        :rtype string
        """
        if enabled:
            self.Enabled = 'Enabled'
        else:
            self.Enabled = 'Disabled'

    def get_trigger_source(self):
        return CRONTAB_TRIGGER


class HttpTriggerData(AbstractTriggerDataModel):
    """
    Http Trigger Data
    :param resource_path  (required) The url path.
    :type resource_path string
    :param method  (required) The http method. eg "GET,HEAD"
    :type method string
    :param auth_type  (required) Authentication type.
    :type auth_type string. eg anonymous | iam
    """

    def __init__(self, resource_path=None, method=None, auth_type=None):
        self.ResourcePath = resource_path
        self.Method = method
        self.AuthType = auth_type

    def get_trigger_source(self):
        return HTTP_TRIGGER


class CdnTriggerData(AbstractTriggerDataModel):
    """
    Cdn Trigger Data
    :param event_type  (required) Cdn event type. The details see
        https://cloud.baidu.com/doc/CFC/s/Kjwvz47o9/#relationconfiguration.
    :type event_type string
    :param domains. Domain list.
    :type domains list of string
    :param remark.
    :type remark string.
    :param status. Enables the trigger.
    :type status bool.
    """

    def __init__(self, event_type=None, domains=None, remark=None, status=False):
        self.EventType = event_type
        self.Domains = domains
        self.Remark = remark
        if status:
            self.Status = 'enabled'
        else:
            self.Status = 'disabled'

    def set_status(self, enabled=False):
        """
        set cdn trigger status

        :param enabled.
        :type enabled bool. Enables the trigger.
        :return
        :rtype string
        """
        if enabled:
            self.Status = 'enabled'
        else:
            self.Status = 'disabled'

    def get_trigger_source(self):
        return CDN_TRIGGER


class BOSTriggerData(AbstractTriggerDataModel):
    """
    BOS Trigger Data
    :param event_type  (required) BOS event type. The details see
        https://cloud.baidu.com/doc/CFC/s/Kjwvz47o9/#relationconfiguration.
    :type event_type list of string
    :param resource. For example, /prefix*suffix  /my.img  /my*img
    :type resource string
    :param name. The name of the trigger that you are creating or updating
    :type name string
    :param status. Enables the trigger.
    :type status bool.
    """

    def __init__(self, bucket=None, event_type=None, resource=None, name=None, status=False):
        self.Resource = resource
        self.EventType = event_type
        self.Name = name
        self.Bucket = bucket
        if status:
            self.Status = 'enabled'
        else:
            self.Status = 'disabled'

    def set_status(self, enabled=False):
        """
        set bos trigger status

        :param enabled.
        :type enabled bool. Enables the trigger.
        :return
        :rtype string
        """
        if enabled:
            self.Status = 'enabled'
        else:
            self.Status = 'disabled'

    def get_trigger_source(self):
        return BOS_TRIGGER + '/' + self.Bucket



# Copyright (c) 2014 Baidu.com, Inc. All Rights Reserved
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file
# except in compliance with the License. You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the
# License is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND,
# either express or implied. See the License for the specific language governing permissions
# and limitations under the License.
"""
This module provides a client class for CFC.
API Reference: https://cloud.baidu.com/doc/CFC/index.html
"""

import copy
import json
import logging
import baidubce
import sys
import time
import traceback
import base64

from baidubce import bce_base_client
from baidubce import utils
from baidubce.auth import bce_v1_signer
from baidubce.http import bce_http_client
from baidubce.http import handler
from baidubce.http import http_content_types
from baidubce.http import http_headers
from baidubce.http import http_methods
from baidubce.services.cfc import cfc_handler
from baidubce.services.cfc import models
from baidubce.exception import BceClientError
from baidubce.exception import BceServerError
from baidubce.utils import required
from baidubce import compat


if compat.PY3:
    from urllib.parse import quote
else:
    from urllib import quote

_logger = logging.getLogger(__name__)


class CfcClient(bce_base_client.BceBaseClient):
    """
    CdnClient
    """
    prefix = '/v1'

    def __init__(self, config=None):
        bce_base_client.BceBaseClient.__init__(self, config)

    def invocations(self, function_name, invocation_type="RequestResponse",
                    log_type="None", body=None, qualifier=None, config=None):
        """
        invoking function

        :param function_name  (required) The cfc function name. You can specify a function name (function_name)
                                or you can specify Baidu Resource Name (BRN) of the function (for example,
                                brn:bce:cfc:bj:account-id:function:function_name). Cfc also allows you
                                to specify a simple BRN (for example, account_id:function_name). The length of BRN is
                                limited to 1 to 140 characters. The function name is limited to 64 characters in length.
        :type function_name string

        :param invocation_type: (required)  Event/RequestResponse/DryRun
        :type invocation_type string

        :param  log_type: None / Tail You can set this optional parameter to Tail in the request only if you
                         specify the InvocationType parameter with value RequestResponse. In this case,CFC
                         returns the base64-encoded last 4 KB of log data produced by your cfc function in
                         the x-bce-log-result header.
        :type log_type string

        :param qualifier Minimum length of 1. Maximum length of 128. You can use function versions or function aliases.
                         If you don't, the default is $LATEST.
        :type qualifier string

        :param body: json

        :param config: None
        :type config: baidubce.BceClientConfiguration

        :return:
        :rtype: httplib.httpresponse
        """
        params = {}
        params["invocationType"] = invocation_type
        params["logType"] = log_type
        if qualifier is not None:
            params["Qualifier"] = qualifier
        if body is None:
            body = {}
        return self._send_request(
            http_methods.POST,
            '/functions/' + function_name + '/invocations',
            body=json.dumps(body),
            params=params,
            config=config, special=True)

    def invoke(self, function_name, invocation_type="RequestResponse",
                    log_type="None", body=None, qualifier=None, config=None):
        """
        invoking function

        :param function_name  (required) The cfc function name. You can specify a function name (function_name)
                                or you can specify Baidu Resource Name (BRN) of the function (for example,
                                brn:bce:cfc:bj:account-id:function:function_name). Cfc also allows you
                                to specify a simple BRN (for example, account_id:function_name). The length of BRN is
                                limited to 1 to 140 characters. The function name is limited to 64 characters in length.
        :type function_name string

        :param invocation_type: (required)  Event/RequestResponse/DryRun
        :type invocation_type string

        :param  log_type: None / Tail You can set this optional parameter to Tail in the request only if you
                         specify the InvocationType parameter with value RequestResponse. In this case,CFC
                         returns the base64-encoded last 4 KB of log data produced by your cfc function in
                         the x-bce-log-result header.
        :type log_type string

        :param qualifier Minimum length of 1. Maximum length of 128. You can use function versions or function aliases.
                         If you don't, the default is $LATEST.
        :type qualifier string

        :param body: json

        :param config: None
        :type config: baidubce.BceClientConfiguration

        :return:
        :rtype: httplib.httpresponse
        """
        return self.invocations(function_name, invocation_type, log_type, body, qualifier, config)

    def create_function(self, function_name, description=None, environment=None,
                        handler=None, memory_size=128, region='bj',
                        zip_file=None, publish=False, run_time='python2',
                        timeout=3, dry_run=False, code_zip_file=None, config=None):
        """
        Create cfc function

        :param function_name  (required)
        :type function_name string

        :param description - A short, user-defined function description. Minimum length of 0. Maximum length of 256.
        :type description string

        :param environment environment's configuration settings.
        :type environment String to string map, [a-zA-Z]([a-zA-Z0-9_])+

        :param handler  (required) Maximum length of 128. [^\s]+
        :type handler string

        :param memory_size  The amount of memory, in MB.The default value is 128 MB. The value must be a multiple of 64
                            MB. From 128M to 3008M. Now it only supports 128M.
        :type memory_size int

        :param region  bj or gz. Now it only supports bj.
        :type region string

        :param zip_file  (required)
        :type zip_file bytes The contents of your zip file containing your deployment package.
                        the contents of the zip file must be base64-encoded.

        :param config: None
        :type config: baidubce.BceClientConfiguration

        :param publish  This boolean parameter can be used to request CFC to create the CFC function and publish
                        a version as an atomic operation.
        :type publish: boolean

        :param run_time: python2 | nodejs6.11
        :type run_time: string

        :param timeout: 1-300 The default is 3 seconds.
        :type timeout: int

        :param code_zip_file: the file path of the zipped code.
        :type code_zip_file: string

        :return:
        :rtype: baidubce.bce_response.BceResponse
        """
        if environment is None:
            environment = {}
        data = {
            'Code': {
                'Publish': publish,
                'DryRun': dry_run
            },
            'Description': description,
            'Region': region,
            'Timeout': timeout,
            'FunctionName': function_name,
            'Handler': handler,
            'Runtime': run_time,
            'MemorySize': memory_size,
            'Environment': {
                'Variables': environment
            }
        }
        if code_zip_file:
            code_file = open(code_zip_file, 'rb')
            code = code_file.read()
            code_base64 = base64.b64encode(code).decode('utf-8')
            data['Code']['ZipFile'] = code_base64
        else:
            data['Code']['ZipFile'] = zip_file
        params = {}
        return self._send_request(
            http_methods.POST,
            '/functions',
            body=json.dumps(data),
            params=params,
            config=config)

    def list_functions(self, function_version=None, page=None, page_size=None,
                       marker=None, max_items=None, config=None):
        """
        List cfc function

        :param function_version
        :type function_version string

        :param page
        :type page int

        :param page_size
        :type page_size int

        :param marker
        :type marker int

        :param max_items
        :type max_items int

        :param config: None
        :type config: baidubce.BceClientConfiguration

        :return:
        :rtype: baidubce.bce_response.BceResponse
        """
        params = {}
        if function_version is not None:
            params['FunctionVersion'] = function_version
        if page is not None:
            params['page'] = page
        if page_size is not None:
            params['pageSize'] = page_size
        if marker is not None:
            params['Marker'] = marker
        if max_items is not None:
            params['MaxItems'] = max_items
        return self._send_request(
            http_methods.GET,
            '/functions/',
            body={},
            params=params,
            config=config)

    def get_function(self, function_name, qualifier=None, config=None):
        """
        get function

        :param function_name  (required) The cfc function name. You can specify a function name (function_name)
                                or you can specify Baidu Resource Name (BRN) of the function (for example,
                                brn:bce:cfc:bj:account-id:function:function_name). Cfc also allows you
                                to specify a simple BRN (for example, account_id:function_name). The length of BRN is
                                limited to 1 to 140 characters. The function name is limited to 64 characters in length.
        :type function_name string

        :param qualifier Minimum length of 1. Maximum length of 128. You can use function versions or function aliases.
                         If you don't, the default is $LATEST.
        :type qualifier string

        :param config: None
        :type config: baidubce.BceClientConfiguration

        :return:
        :rtype: baidubce.bce_response.BceResponse
        """
        params = {}
        if qualifier is not None:
            params["Qualifier"] = qualifier
        return self._send_request(
            http_methods.GET,
            '/functions/' + function_name,
            body={},
            params=params,
            config=config)

    def delete_function(self, function_name, qualifier=None, config=None):
        """
        delete_function

        :param function_name  (required) The cfc function name. You can specify a function name (function_name)
                                or you can specify Baidu Resource Name (BRN) of the function (for example,
                                brn:bce:cfc:bj:account-id:function:function_name). Cfc also allows you
                                to specify a simple BRN (for example, account_id:function_name). The length of BRN is
                                limited to 1 to 140 characters. The function name is limited to 64 characters in length.
        :type function_name string

        :param qualifier Minimum length of 1. Maximum length of 128. You can use function versions or function aliases.
                         If you don't, the default is $LATEST.
        :type qualifier string

        :param config: None
        :type config: baidubce.BceClientConfiguration

        :return:
        :rtype: baidubce.bce_response.BceResponse
        """
        params = {}
        if qualifier is not None:
            params["Qualifier"] = qualifier
        return self._send_request(
            http_methods.DELETE,
            '/functions/' + function_name,
            body={},
            params={},
            config=config)

    def update_function_code(self, function_name, zip_file=None,
                             publish=None, dry_run=None, config=None):
        """
        update_function_code

        :param function_name  (required) The cfc function name. You can specify a function name (function_name)
                                or you can specify Baidu Resource Name (BRN) of the function (for example,
                                brn:bce:cfc:bj:account-id:function:function_name). Cfc also allows you
                                to specify a simple BRN (for example, account_id:function_name). The length of BRN is
                                limited to 1 to 140 characters. The function name is limited to 64 characters in length.
        :type function_name string

        :param zip_file
        :type zip_file bytes The contents of your zip file containing your deployment package.
                        the contents of the zip file must be base64-encoded.
        :param publish  This boolean parameter can be used to request CFC to create the CFC function and publish
                        a version as an atomic operation.
        :type publish: boolean

        :param dry_run  This boolean parameter can be used to test your request to CFC to update the function and
                        publish a version as an atomic operation. It will do all necessary computation and validation
                        of your code but will not upload it or a publish a version. Each time this operation is invoked,
                        the CodeSha256 hash value of the provided code will also be computed and returned in the
                        response.
        :type dry_run: boolean

        :param config: None
        :type config: baidubce.BceClientConfiguration

        :return:
        :rtype: baidubce.bce_response.BceResponse
        """
        body = {}
        if zip_file is not None:
            body["ZipFile"] = zip_file
        if publish is not None:
            body["Publish"] = publish
        if dry_run is not None:
            body["DryRun"] = dry_run

        return self._send_request(
            http_methods.PUT,
            '/functions/' + function_name + '/code',
            body=json.dumps(body),
            params={},
            config=config)

    def get_function_configuration(self, function_name, qualifier=None, config=None):
        """
        get_function_configuration
        :param function_name  (required) The cfc function name. You can specify a function name (function_name)
                                or you can specify Baidu Resource Name (BRN) of the function (for example,
                                brn:bce:cfc:bj:account-id:function:function_name). Cfc also allows you
                                to specify a simple BRN (for example, account_id:function_name). The length of BRN is
                                limited to 1 to 140 characters. The function name is limited to 64 characters in length.
        :type function_name string

        :param qualifier Minimum length of 1. Maximum length of 128. You can use function versions or function aliases.
                         If you don't, the default is $LATEST.
        :type qualifier string

        :param config: None
        :type config: baidubce.BceClientConfiguration

        :return:
        :rtype: baidubce.bce_response.BceResponse
        """
        params = {}
        if qualifier is not None:
            params["Qualifier"] = qualifier

        return self._send_request(
            http_methods.GET,
            '/functions/' + function_name + "/configuration",
            body={},
            params={},
            config=config)

    def update_function_configuration(self, function_name, description=None, environment=None,
                                      handler=None, run_time=None,
                                      timeout=None, config=None):
        """
        update_function_configuration
        :param function_name  (required)
        :type function_name string

        :param description - A short, user-defined function description. Minimum length of 0. Maximum length of 256.
        :type description string

        :param environment environment's configuration settings.
        :type environment String to string map, [a-zA-Z]([a-zA-Z0-9_])+

        :param handler  (required) Maximum length of 128. [^\s]+
        :type handler string

        :param run_time: python2 | nodejs6.11
        :type run_time: string

        :param timeout: 1-300 The default is 3 seconds.
        :type timeout: int

        :param config: None
        :type config: baidubce.BceClientConfiguration

        :return:
        :rtype: baidubce.bce_response.BceResponse
        """
        data = {}
        if description is not None:
            data["Description"] = description
        if environment is not None:
            data["Environment"] = {"Variables": environment}
        if handler is not None:
            data["Handler"] = handler
        if run_time is not None:
            data["Runtime"] = run_time
        if timeout is not None:
            data["Timeout"] = timeout
        return self._send_request(
            http_methods.PUT,
            '/functions/' + function_name + '/configuration',
            body=json.dumps(data),
            params={},
            config=config)

    def list_versions_by_function(self, function_name, marker=None, max_items=None, config=None):
        """
        list_versions_by_function
        :param function_name  (required) The cfc function name. You can specify a function name (function_name)
                                or you can specify Baidu Resource Name (BRN) of the function (for example,
                                brn:bce:cfc:bj:account-id:function:function_name). Cfc also allows you
                                to specify a simple BRN (for example, account_id:function_name). The length of BRN is
                                limited to 1 to 140 characters. The function name is limited to 64 characters in length.
        :type function_name string

                :param marker
        :type marker int

        :param max_items
        :type max_items int

        :param config: None
        :type config: baidubce.BceClientConfiguration

        :return:
        :rtype: baidubce.bce_response.BceResponse
        """
        params = {}
        if marker is not None:
            params["Marker"] = marker
        if max_items is not None:
            params["MaxItems"] = max_items
        return self._send_request(
            http_methods.GET,
            '/functions/' + function_name + "/versions",
            body={},
            params=params,
            config=config)

    def publish_version(self, function_name, description=None, config=None):
        """
        publish_version
        :param function_name  (required) The cfc function name. You can specify a function name (function_name)
                                or you can specify Baidu Resource Name (BRN) of the function (for example,
                                brn:bce:cfc:bj:account-id:function:function_name). Cfc also allows you
                                to specify a simple BRN (for example, account_id:function_name). The length of BRN is
                                limited to 1 to 140 characters. The function name is limited to 64 characters in length.
        :type function_name string

        :param description  A short, user-defined function publish description. Minimum length of 0. Maximum length of
                            256.
        :type description string

        :param config: None
        :type config: baidubce.BceClientConfiguration

        :return:
        :rtype: baidubce.bce_response.BceResponse
        """
        params = {}
        if description is not None:
            params["Description"] = description

        return self._send_request(
            http_methods.POST,
            '/functions/' + function_name + "/versions",
            body={},
            params=params,
            config=config)

    def list_aliases(self, function_name, function_version=None,
                     marker=None, max_items=None, config=None):
        """
        list_aliases
        :param function_name  (required) The cfc function name. You can specify a function name (function_name)
                                or you can specify Baidu Resource Name (BRN) of the function (for example,
                                brn:bce:cfc:bj:account-id:function:function_name). Cfc also allows you
                                to specify a simple BRN (for example, account_id:function_name). The length of BRN is
                                limited to 1 to 140 characters. The function name is limited to 64 characters in length.
        :type function_name string

        :param function_version
        :type function_version string

        :param marker
        :type marker int

        :param max_items
        :type max_items int

        :param config: None
        :type config: baidubce.BceClientConfiguration

        :return:
        :rtype: baidubce.bce_response.BceResponse
        """
        params = {}
        if function_version is not None:
            params["FunctionVersion"] = function_version
        if marker is not None:
            params["Marker"] = marker
        if max_items is not None:
            params["MaxItems"] = max_items

        return self._send_request(
            http_methods.GET,
            '/functions/' + function_name + "/aliases",
            body={},
            params=params,
            config=config)

    def create_alias(self, function_name, function_version=None,
                     name=None, description=None, config=None):
        """
        create_alias
        :param function_name  (required) The cfc function name. You can specify a function name (function_name)
                                or you can specify Baidu Resource Name (BRN) of the function (for example,
                                brn:bce:cfc:bj:account-id:function:function_name). Cfc also allows you
                                to specify a simple BRN (for example, account_id:function_name). The length of BRN is
                                limited to 1 to 140 characters. The function name is limited to 64 characters in length.
        :type function_name string

        :param name alias name
        :type name string

        :param function_version function version for which you are creating the alias. (\$LATEST|[0-9]+)
        :type function_version string

        :param config: None
        :type config: baidubce.BceClientConfiguration

        :param description   Description of the alias.
        :type description string

        :return:
        :rtype: baidubce.bce_response.BceResponse
        """
        data = {}
        if description is not None:
            data["Description"] = description
        if function_version is not None:
            data["FunctionVersion"] = function_version
        if name is not None:
            data["Name"] = name

        return self._send_request(
            http_methods.POST,
            '/functions/' + function_name + "/aliases",
            body=json.dumps(data),
            params={},
            config=config)

    def get_alias(self, function_name, alias_name, config=None):
        """
        get_alias
        :param function_name  (required) The cfc function name. You can specify a function name (function_name)
                                or you can specify Baidu Resource Name (BRN) of the function (for example,
                                brn:bce:cfc:bj:account-id:function:function_name). Cfc also allows you
                                to specify a simple BRN (for example, account_id:function_name). The length of BRN is
                                limited to 1 to 140 characters. The function name is limited to 64 characters in length.
        :type function_name string

        :param alias_name alias name
        :type alias_name string

        :param config: None
        :type config: baidubce.BceClientConfiguration

        :return:
        :rtype: baidubce.bce_response.BceResponse
        """
        return self._send_request(
            http_methods.GET,
            '/functions/' + function_name + "/aliases/" + alias_name,
            body={},
            params={},
            config=config)

    def update_alias(self, function_name, alias_name, function_version=None,
                     description=None, config=None):
        """
        update_alias
        :param function_name  (required) The cfc function name. You can specify a function name (function_name)
                                or you can specify Baidu Resource Name (BRN) of the function (for example,
                                brn:bce:cfc:bj:account-id:function:function_name). Cfc also allows you
                                to specify a simple BRN (for example, account_id:function_name). The length of BRN is
                                limited to 1 to 140 characters. The function name is limited to 64 characters in length.
        :type function_name string

        :param alias_name alias name
        :type alias_name string

        :param function_version function version for which you are update the alias. (\$LATEST|[0-9]+)
        :type function_version string

        :param description   Description of the alias.
        :type description string

        :param config: None
        :type config: baidubce.BceClientConfiguration

        :return:
        :rtype: baidubce.bce_response.BceResponse
        """
        data = {}
        if description is not None:
            data["Description"] = description
        if function_version is not None:
            data["FunctionVersion"] = function_version

        return self._send_request(
            http_methods.PUT,
            '/functions/' + function_name + "/aliases/" + alias_name,
            body=json.dumps(data),
            params={},
            config=config)

    def delete_alias(self, function_name, alias_name, config=None):
        """
        delete_alias
        :param function_name  (required) The cfc function name. You can specify a function name (function_name)
                                or you can specify Baidu Resource Name (BRN) of the function (for example,
                                brn:bce:cfc:bj:account-id:function:function_name). Cfc also allows you
                                to specify a simple BRN (for example, account_id:function_name). The length of BRN is
                                limited to 1 to 140 characters. The function name is limited to 64 characters in length.
        :type function_name string

        :param alias_name alias name
        :type alias_name string

        :param config: None
        :type config: baidubce.BceClientConfiguration

        :return:
        :rtype: baidubce.bce_response.BceResponse
        """
        return self._send_request(
            http_methods.DELETE,
            '/functions/' + function_name + "/aliases/" + alias_name,
            body={},
            params={},
            config=config)

    def list_triggers(self, function_brn, config=None):
        """
        list_triggers
        :param function_brn  (required) The cfc function brn.
        :type function_brn string

        :param config: None
        :type config: baidubce.BceClientConfiguration

        :return: message result as following format
            {
              "Relation": [
                {
                  "RelationId": "brn:bce:cfc-http-trigger:bj:cd64f99c69d7c404b61de0a4f1865834:b8542048977633ad0a867aefc33fd32a/cfc/GET/cfc/docs",
                  "Sid": "cfc-dfbd5359-1afb-49e7-81d7-16554056c79d",
                  "Source": "cfc-http-trigger/v1/CFCAPI",
                  "Target": "brn:bce:cfc:bj:cd64f99c69d7c404b61de0a4f1865834:function:helloCFCDocs:$LATEST",
                  "Data": {
                    "AuthType": "anonymous",
                    "Brn": "brn:bce:cfc-http-trigger:bj:cd64f99c69d7c404b61de0a4f1865834:b8542048977633ad0a867aefc33fd32a/cfc/GET/cfc/docs",
                    "EndpointPrefix": "https://6ewfn1337kndc.cfc-execute.bj.baidubce.com",
                    "Method": "GET",
                    "ResourcePath": "/cfc/docs"
                  }
                },
                {
                  "RelationId": "brn:bce:cfc-http-trigger:bj:cd64f99c69d7c404b61de0a4f1865834:b8542048977633ad0a867aefc33fd32a/cfc/POST,PUT/cfc/docs/edit",
                  "Sid": "cfc-dfbd5359-1afb-49e7-81d7-16554056c79d",
                  "Source": "cfc-http-trigger/v1/CFCAPI",
                  "Target": "brn:bce:cfc:bj:cd64f99c69d7c404b61de0a4f1865834:function:helloCFCDocs:$LATEST",
                  "Data": {
                    "AuthType": "anonymous",
                    "Brn": "brn:bce:cfc-http-trigger:bj:cd64f99c69d7c404b61de0a4f1865834:b8542048977633ad0a867aefc33fd32a/cfc/POST,PUT/cfc/docs/edit",
                    "EndpointPrefix": "https://6ewfn1337kndc.cfc-execute.bj.baidubce.com",
                    "Method": "POST,PUT",
                    "ResourcePath": "/cfc/docs/edit"
                  }
                }
              ]
            }
        :rtype: baidubce.bce_response.BceResponse
        """
        params = {}
        if function_brn is not None:
            params["FunctionBrn"] = function_brn
        return self._send_request(
            http_methods.GET,
            '/relation',
            body={},
            params=params,
            config=config)

    def create_trigger(self, function_brn, source=None, trigger_data=None, config=None):
        """
        create_trigger
        :param function_brn  (required) The cfc function brn.
        :type function_brn string

        :param source. The trigger source type. For example, models.CRONTAB_TRIGGER
        :type source string

        :param trigger_data The trigger data. You can get details from
            https://cloud.baidu.com/doc/CFC/s/Kjwvz47o9/#relationconfiguration
        :type trigger_data models.AbstractTriggerDateModel

        :param config: None
        :type config: baidubce.BceClientConfiguration

        :return:  message result as following format
            {
                "Relation": {
                    "Sid": "cfc-c53bef4e-2cac-4bc6-84c1-2d5ec2f8bec2",
                    "RelationId" : "00457f0b-20d8-4f3d-8555-c2u121f38313",
                    "Source": "string",
                    "Target": "brn:bce:cfc:bj:640c8817bd1de2928d47256dd0620ce5:function:test:$LATEST",
                    "Data": {
                        "EventType":["PutObject", "PostObject"],
                        "Prefix":"images/",
                        "Suffix":".jpg",
                        "Status":"enabled"
                    }
                }
            }
        :rtype: baidubce.bce_response.BceResponse
        """
        data = {}
        if function_brn is not None:
            data["Target"] = function_brn
        if source is not None:
            data["Source"] = source
        if trigger_data is not None:
            if isinstance(trigger_data, models.AbstractTriggerDataModel):
                data["Source"] = trigger_data.get_trigger_source()
                data["Data"] = trigger_data.serialize()
            else:
                data["Data"] = trigger_data
        return self._send_request(
            http_methods.POST,
            '/relation',
            body=json.dumps(data),
            params={},
            config=config)

    def update_trigger(self, function_brn, relation_id, trigger_data, source=None, config=None):
        """
        create_trigger
        :param function_brn  (required) The cfc function brn.
        :type function_brn string

        :param source  (required) The trigger source type. For example, models.CRONTAB_TRIGGER
        :type source string

        :param relation_id  (required) The relation_id.
        :type relation_id string

        :param trigger_data The trigger data. You can get details from
            https://cloud.baidu.com/doc/CFC/s/Kjwvz47o9/#relationconfiguration
        :type trigger_data models.AbstractTriggerDateModel

        :param config: None
        :type config: baidubce.BceClientConfiguration

        :return: message result as following format
            {
                "Relation": {
                    "Sid": "cfc-c53bef4e-2cac-4bc6-84c1-2d5ec2f8bec2",
                    "RelationId" : "00457f0b-20d8-4f3d-8555-c2u121f38313",
                    "Source": "string",
                    "Target": "brn:bce:cfc:bj:640c8817bd1de2928d47256dd0620ce5:function:test:$LATEST",
                    "Data": {
                        "EventType":["PutObject", "PostObject"],
                        "Prefix":"images/",
                        "Suffix":".jpg",
                        "Status":"enabled"
                    }
                }
            }
        :rtype: baidubce.bce_response.BceResponse
        """
        data = {}
        if function_brn is not None:
            data["Target"] = function_brn
        if source is not None:
            data["Source"] = source
        if trigger_data is not None:
            if isinstance(trigger_data, models.AbstractTriggerDataModel):
                data["Source"] = trigger_data.get_trigger_source()
                data["Data"] = trigger_data.serialize()
            else:
                data["Data"] = trigger_data
        if relation_id is not None:
            data["RelationId"] = relation_id

        return self._send_request(
            http_methods.PUT,
            '/relation',
            body=json.dumps(data),
            params={},
            config=config)

    def delete_trigger(self, function_brn, source, relation_id, config=None):
        """
        delete_trigger
        :param function_brn  (required) The cfc function brn.
        :type function_brn string

        :param source  (required) The trigger source type.
        :type source string

        :param relation_id  (required) The relation_id.
        :type relation_id string

        :param config: None
        :type config: baidubce.BceClientConfiguration

        :return:
        :rtype: baidubce.bce_response.BceResponse
        """
        params = {}
        if function_brn is not None:
            params["Target"] = function_brn
        if source is not None:
            params["Source"] = source
        if relation_id is not None:
            params["RelationId"] = relation_id
        return self._send_request(
            http_methods.DELETE,
            '/relation',
            body={},
            params=params,
            config=config)

    @staticmethod
    def _encode_function_name(self, function_name):
        return ''

    @staticmethod
    def _merge_config(self, config):
        if config is None:
            return self.config
        else:
            new_config = copy.copy(self.config)
            new_config.merge_non_none_values(config)
            return new_config

    def _send_request(
            self, http_method, path,
            body=None, headers=None, params=None,
            config=None,
            body_parser=None, special=False):
        config = self._merge_config(self, config)
        if body_parser is None:
            body_parser = cfc_handler.parse_json
        headers = headers or {}
        headers[http_headers.CONTENT_TYPE] = http_content_types.JSON
        if config.security_token is not None:
            headers = headers or {}
            headers[http_headers.STS_SECURITY_TOKEN] = config.security_token
        return self.send_request(config, bce_v1_signer.sign,
                                 [cfc_handler.parse_error, body_parser],
                                 http_method, CfcClient.prefix + path,
                                 body, headers, params, special)

    def send_request(
            self,
            config,
            sign_function,
            response_handler_functions,
            http_method, path, body, headers, params, special=False):
        """
        Send request to BCE services.
        :param config
        :type config: baidubce.BceClientConfiguration
        :param sign_function:
        :param response_handler_functions:
        :type response_handler_functions: list
        :param request:
        :type request: baidubce.internal.InternalRequest
        :return:
        :rtype: baidubce.BceResponse
        """
        t = int(time.time())
        _logger.debug(b'%s request start: %s %s, %s, %s, %d',
                      http_method, path, headers, params, body, t)
        headers = headers or {}
        if config.security_token is not None:
            headers[http_headers.STS_SECURITY_TOKEN] = config.security_token
        headers_to_sign = [b"host",
                           b"content-length",
                           b"content-type"]

        should_get_new_date = False
        if http_headers.BCE_DATE not in headers:
            should_get_new_date = True
        headers[http_headers.HOST] = config.endpoint

        for k in headers:
            k_lower = k.strip().lower()
            if k_lower.startswith(http_headers.BCE_PREFIX):
                headers_to_sign.append(k_lower)
        user_agent = 'bce-sdk-python/%s/%s/%s' % (
            compat.convert_to_string(baidubce.SDK_VERSION), sys.version, sys.platform)
        user_agent = user_agent.replace('\n', '')
        user_agent = compat.convert_to_bytes(user_agent)
        headers[http_headers.USER_AGENT] = user_agent
        body = compat.convert_to_bytes(body)
        if not body:
            headers[http_headers.CONTENT_LENGTH] = 0
        elif isinstance(body, bytes):
            headers[http_headers.CONTENT_LENGTH] = len(body)
        elif http_headers.CONTENT_LENGTH not in headers:
            raise ValueError(b'No %s is specified.' % http_headers.CONTENT_LENGTH)

        # store the offset of fp body
        offset = None
        if hasattr(body, "tell") and hasattr(body, "seek"):
            offset = body.tell()

        protocol, host, port = utils.parse_host_port(config.endpoint, config.protocol)
        path = quote(path)

        headers[http_headers.HOST] = host
        if port != config.protocol.default_port:
            headers[http_headers.HOST] += b':' + compat.convert_to_bytes(port)
        path = compat.convert_to_bytes(path)
        headers[http_headers.AUTHORIZATION] = sign_function(
            config.credentials, http_method, path, headers, params, headers_to_sign=headers_to_sign)
        encoded_params = utils.get_canonical_querystring(params, False)
        if len(encoded_params) > 0:
            uri = path + b'?' + encoded_params
        else:
            uri = path
        bce_http_client.check_headers(headers)

        retries_attempted = 0
        errors = []
        while True:
            conn = None
            try:
                if should_get_new_date is True:
                    headers[http_headers.BCE_DATE] = utils.get_canonical_time()
                    headers_to_sign.append(http_headers.BCE_DATE)

                headers[http_headers.AUTHORIZATION] = sign_function(
                    config.credentials, http_method, path, headers, params, headers_to_sign=headers_to_sign)
                if retries_attempted > 0 and offset is not None:
                    body.seek(offset)

                conn = bce_http_client._get_connection(protocol, host,
                                                       port, config.connection_timeout_in_mills)

                _logger.debug('request args:method=%s, uri=%s, headers=%s,patams=%s, body=%s',
                              http_method, uri, headers, params, body)

                http_response = bce_http_client._send_http_request(
                    conn, http_method, uri, headers, body, config.send_buf_size)

                headers_list = http_response.getheaders()

                # on py3 ,values of headers_list is decoded with ios-8859-1 from
                # utf-8 binary bytes

                # headers_list[*][0] is lowercase on py2
                # headers_list[*][0] is raw value py3
                if compat.PY3 and isinstance(headers_list, list):
                    temp_heads = []
                    for k, v in headers_list:
                        k = k.encode('latin-1').decode('utf-8')
                        v = v.encode('latin-1').decode('utf-8')
                        k = k.lower()
                        temp_heads.append((k, v))
                    headers_list = temp_heads

                _logger.debug(
                    'request return: status=%d, headers=%s' % (http_response.status, headers_list))
                # cfc invoke return doesn't have to be json
                if special:
                    return http_response

                response = bce_http_client.BceResponse()
                response.set_metadata_from_headers(dict(headers_list))
                for handler_function in response_handler_functions:
                    if handler_function(http_response, response):
                        break

                return response
            except Exception as e:
                if conn is not None:
                    conn.close()
                # insert ">>>>" before all trace back lines and then save it
                errors.append('\n'.join('>>>>' + line for line in traceback.format_exc().splitlines()))
                if config.retry_policy.should_retry(e, retries_attempted):
                    delay_in_millis = config.retry_policy.get_delay_before_next_retry_in_millis(
                        e, retries_attempted)
                    time.sleep(delay_in_millis / 1000.0)
                else:
                    raise bce_http_client.BceHttpClientError('Unable to execute HTTP request. '
                                                             'Retried %d times. All trace backs:\n'
                                                             '%s' % (retries_attempted,
                                                                     '\n'.join(errors)), e)
        retries_attempted += 1



# -*- coding: utf-8 -*-

# Copyright (c) 2014 Baidu.com, Inc. All Rights Reserved
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions
# and limitations under the License.

"""
This module provides a client class for EIP group.
"""

import copy
import json
import logging
import uuid

from baidubce import utils
from baidubce import bce_base_client
from baidubce.auth import bce_v1_signer
from baidubce.http import bce_http_client
from baidubce.http import handler
from baidubce.http import http_methods
from baidubce.services.eip import eip_group_model
from baidubce.utils import required

_logger = logging.getLogger(__name__)


default_billing_to_purchase_created = eip_group_model.Billing('Prepaid')
default_billing_to_purchase_reserved = eip_group_model.Billing()


class EipGroupClient(bce_base_client.BceBaseClient):
    """
    EIP group sdk client
    """
    version = b'/v1'
    prefix = b'/eipgroup'

    def __init__(self, config=None):
        bce_base_client.BceBaseClient.__init__(self, config)

    def _merge_config(self, config=None):
        """
        :param config:
        :type config: baidubce.BceClientConfiguration
        :return:
        """
        if config is None:
            return self.config
        else:
            new_config = copy.copy(self.config)
            new_config.merge_non_none_values(config)
            return new_config

    def _send_request(self, http_method, path,
                      body=None, headers=None, params=None,
                      config=None, body_parser=None):
        config = self._merge_config(config)
        if body_parser is None:
            body_parser = handler.parse_json
        if headers is None:
            headers = {b'Accept': b'*/*',
                       b'Content-Type': b'application/json;charset=utf-8'}
        return bce_http_client.send_request(
            config, bce_v1_signer.sign, [handler.parse_error, body_parser],
            http_method, path, body, headers, params)

    @required(eip_count=int,
              bandwidth_in_mbps=int)
    def create_eip_group(self, eip_count, bandwidth_in_mbps,
                         name=None, client_token=None,
                         billing=None, config=None):
        """
        Create a shared bandwidth EIP group with specified options.
        Real-name authentication is required before creating EIP groups.
        Only prepaid EIP groups is supported.

        :param client_token:
            An ASCII string whose length is less than 64.
            The request will be idempotent if client token is provided.
        :type client_token: string

        :param eip_count:
            Numbers of EIP addresses in the EIP group.
            The minimum number of public IP addresses is two,
            and the maximum number multiplies 5Mbps mustn't exceed the
            total amount of shared bandwidth package.
        :type eip_count: int

        :param bandwidth_in_mbps:
            Public Internet bandwidth in unit Mbps. For Prepaid EIP groups,
            this value must be integer between 10 and 200.
        :type bandwidth_in_mbps: int

        :param billing:
            Billing information.
        :type billing: eip_group_model.Billing

        :param name:
            The name of EIP group that will be created.
            The name, beginning with letter, should have the length between
            1 and 65 bytes, and could contain alphabets, numbers or '-_/.'.
            If not specified, the service will generate it automatically.
        :type name: string

        :param config:
        :type config: baidubce.BceClientConfiguration

        :return:
        :rtype baidubce.bce_response.BceResponse
        """
        path = self._get_path()
        if client_token is None:
            client_token = generate_client_token()
        params = {b'clientToken': client_token}
        if billing is None:
            billing = default_billing_to_purchase_created
        body = {
            'eipCount': eip_count,
            'bandwidthInMbps': bandwidth_in_mbps,
            'billing': billing.__dict__
        }
        if name is not None:
            body['name'] = name
        return self._send_request(http_methods.POST,
                                  path, body=json.dumps(body),
                                  params=params, config=config)

    def list_eip_groups(self, id=None, name=None, status=None,
                        marker=None, max_keys=None, config=None):
        """
        Return a list of EIP groups, according to the ID,
        name or status of EIP group. If not specified,
        returns a full list of EIP groups in VPC.

        :param id:
            The id of specified EIP group.
        :type id: string

        :param name:
            The name of specified EIP group.
        :type name: string

        :param status:
            The status of specified EIP group.
        :type status: string

        :param marker:
            The optional parameter marker specified in the original
            request to specify where in the results to begin listing.
            Together with the marker, specifies the list result which
            listing should begin. If the marker is not specified,
            the list result will listing from the first one.
        :type marker: string

        :param max_keys:
            The optional parameter to specifies the max number of
            list result to return.
            The default value is 1000.
        :type max_keys: int

        :param config:
        :type config: baidubce.BceClientConfiguration

        :return:
        :rtype baidubce.bce_response.BceResponse
        """
        path = self._get_path()
        params = {}
        if id is not None:
            params[b'id'] = id
        if name is not None:
            params[b'name'] = name
        if status is not None:
            params[b'status'] = status
        if marker is not None:
            params[b'marker'] = marker
        if max_keys is not None:
            params[b'maxKeys'] = max_keys
        return self._send_request(http_methods.GET, path,
                                  params=params, config=config)

    @required(id=(bytes, str))
    def get_eip_group(self, id, config=None):
        """
        Get the detail information of specified EIP group.

        :param id:
            The id of specified EIP group.
        :type id: string

        :param config:
        :type config: baidubce.BceClientConfiguration

        :return:
        :rtype baidubce.bce_response.BceResponse
        """
        path = utils.append_uri(self._get_path(), id)
        return self._send_request(http_methods.GET, path, config=config)

    @required(id=(bytes, str), name=(bytes, str))
    def update_eip_group(self, id, name, client_token=None, config=None):
        """
        Update the name of specified EIP group.

        :param id:
            The id of specified EIP group.
        :type id: string

        :param name:
            The new name of the EIP group
        :type name: string

        :param client_token:
            An ASCII string whose length is less than 64.
            The request will be idempotent if clientToken is provided.
            If the clientToken is not specified by user,
            a random String generated by default algorithm will be used.
        :type client_token: string

        :param config:
        :type config: baidubce.BceClientConfiguration

        :return:
        :rtype baidubce.bce_response.BceResponse
        """
        path = utils.append_uri(self._get_path(), id)
        if client_token is None:
            client_token = generate_client_token()
        params = {
            b'update': None,
            b'clientToken': client_token
        }
        body = {
            'name': name
        }
        return self._send_request(http_methods.PUT,
                                  path, body=json.dumps(body),
                                  params=params, config=config)

    @required(id=(bytes, str), bandwidth_in_mbps=int)
    def resize_eip_group_bandwidth(self, id, bandwidth_in_mbps,
                                   client_token=None, config=None):
        """
        Resize the bandwidth of a specified EIP group.

        :param id:
            The id of specified EIP group.
        :type id: string

        :param client_token:
            An ASCII string whose length is less than 64.
            The request will be idempotent if clientToken is provided.
            If the clientToken is not specified by user,
            a random String generated by default algorithm will be used.
        :type client_token: string

        :param bandwidth_in_mbps:
            The new bandwidth of EIP group.
            For prepaid EIP groups, this value must be integer
            between 10 and 200.
        :type bandwidth_in_mbps: int

        :param config:
        :type config: baidubce.BceClientConfiguration

        :return:
        :rtype baidubce.bce_response.BceResponse
        """
        path = utils.append_uri(self._get_path(), id)
        if client_token is None:
            client_token = generate_client_token()
        params = {
            b'resize': None,
            b'clientToken': client_token
        }
        body = {
            'bandwidthInMbps': bandwidth_in_mbps
        }
        return self._send_request(http_methods.PUT,
                                  path, body=json.dumps(body),
                                  params=params, config=config)

    @required(id=(bytes, str), eip_add_count=int)
    def resize_eip_group_count(self, id, eip_add_count,
                               client_token=None, config=None):
        """
        Resize the EIP count of a specified EIP group.

        :param id:
            The id of specified EIP group.
        :type id: string

        :param client_token:
            An ASCII string whose length is less than 64.
            The request will be idempotent if clientToken is provided.
            If the clientToken is not specified by user,
            a random String generated by default algorithm will be used.
        :type client_token: string

        :param eip_add_count:
            The increase number of EIP addresses in the EIP group.
            This value must larger than zero, and the maximum number multiplies
            5Mbps mustn't exceed the total amount of shared bandwidth package.
        :type eip_add_count: int

        :param config:
        :type config: baidubce.BceClientConfiguration

        :return:
        :rtype baidubce.bce_response.BceResponse
        """
        path = utils.append_uri(self._get_path(), id)
        if client_token is None:
            client_token = generate_client_token()
        params = {
            b'resize': None,
            b'clientToken': client_token
        }
        body = {
            'eipAddCount': eip_add_count
        }
        return self._send_request(http_methods.PUT,
                                  path, body=json.dumps(body),
                                  params=params, config=config)

    @required(id=(bytes, str))
    def purchase_reserved_eip_group(self, id, client_token=None,
                                    billing=None, config=None):
        """
        Renew specified EIP group.
        EIP groups cannot can not be renewed during resizing process.

        :param id:
            The id of EIP group.
        :type id: string

        :param client_token:
            An ASCII string whose length is less than 64.
            The request will be idempotent if clientToken is provided.
            If the clientToken is not specified by user,
            a random String generated by default algorithm will be used.
        :type client_token: string

        :param billing:
            Billing information.
        :type billing: eip_group_model.Billing

        :param config:
        :type config: baidubce.BceClientConfiguration

        :return:
        :rtype baidubce.bce_response.BceResponse
        """
        path = utils.append_uri(self._get_path(), id)
        if client_token is None:
            client_token = generate_client_token()
        params = {
            b'purchaseReserved': None,
            b'clientToken': client_token
        }
        if billing is None:
            billing = default_billing_to_purchase_reserved
        body = {
            'billing': billing.__dict__
        }
        return self._send_request(http_methods.PUT,
                                  path, body=json.dumps(body),
                                  params=params, config=config)

    @staticmethod
    def _get_path(prefix=None):
        """
        :type prefix: string
        :param prefix: path prefix
        """
        if prefix is None:
            prefix = EipGroupClient.prefix
        return utils.append_uri(EipGroupClient.version, prefix)


def generate_client_token_by_uuid():
    """
    The default method to generate the random string for client_token
    if the optional parameter client_token is not specified by the user.

    :return:
    :rtype string
    """
    return str(uuid.uuid4())


generate_client_token = generate_client_token_by_uuid



# -*- coding: utf-8 -*-

# Copyright (c) 2020 Baidu.com, Inc. All Rights Reserved
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy
#  of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions
# and limitations under the License.

"""
This module provides a client class for EIP.
"""

import copy
import json
import logging
import uuid

from baidubce import utils
from baidubce.auth import bce_v1_signer
from baidubce.bce_base_client import BceBaseClient
from baidubce.http import bce_http_client
from baidubce.http import handler
from baidubce.http import http_methods
from baidubce.utils import required

_logger = logging.getLogger(__name__)


class EipBpClient(BceBaseClient):
    """
    EIP_BP sdk client
    """
    version = b'/v1'
    prefix = b'/eipbp'

    def __init__(self, config=None):
        """
        :type config: baidubce.BceClientConfiguration
        :param config:
        """
        BceBaseClient.__init__(self, config)

    @required(bandwidth_in_mbps=int)
    def create_eip_bp(self, eip, eip_group_Id, bandwidth_in_mbps, name=None, autoReleaseTime=None,
                      client_token=None, config=None):
        """
        Create an eip_bp with the specified options.

        :type eip: string
        :param eip: the eip address that the eip_bp attach.
        Param "eip" and "eip_group_Id" will has only one param take effect.

        :type eip_group_Id: string
        :param eip_group_Id: the eipgroupId that the eip_bp attach.
        Param "eip" and "eip_group_Id" will has only one param take effect.

        :type bandwidth_in_mbps: int
        :param bandwidth_in_mbps: the bandwidth of eip_bp in Mbps.

        :type name: string
        :param name:the name of eipbp. The optional parameter.

        :type auto_release_time: string (UTC format,like yyyy:mm:ddThh:mm:ssZ)
        :param autoReleaseTime: the autoReleaseTime of eipbp. The optional parameter.

        :type client_token: string
        :param client_token: if the clientToken is not specified by the user,
         a random string generated by default algorithm will be used.

        :type config: baidubce.BceClientConfiguration
        :param config:

        :return: created eip_bp id, for example,{"id":"bw-xxxxxxxx"}
        """
        body = {
            'bandwidthInMbps': bandwidth_in_mbps
        }
        if eip is None and eip_group_Id is None:
            raise ValueError(b"eip and eip_group_Id can't be none at the same time.")

        if eip is not None:
            body['eip'] = eip
        if eip_group_Id is not None:
            body['eipGroupId'] = eip_group_Id
        if name is not None:
            body['name'] = name
        if autoReleaseTime is not None:
            body['autoReleaseTime'] = autoReleaseTime

        path = self._get_path()
        if client_token is None:
            client_token = self._generate_default_client_token()
        params = {
            b'clientToken': client_token
        }

        return self._send_request(http_methods.POST, path,
                                  body=json.dumps(body), params=params,
                                  config=config)

    @required(id=str, new_bandwidth_in_mbps=int)
    def resize_eip_bp(self, id, new_bandwidth_in_mbps, client_token=None,
                      config=None):
        """
        Resizing eip_bp

        :type id: string
        :param id: eip_bp's id to be resized

        :type new_bandwidth_in_mbps: int
        :param new_bandwidth_in_mbps: specify new bandwidth in Mbps for eip_bp

        :type client_token: string
        :param client_token: if the clientToken is not specified by the user,
         a random string generated by default algorithm will be used.

        :type config: baidubce.BceClientConfiguration
        :param config:

        :return: BceResponse
        """
        body = {
            'bandwidthInMbps': new_bandwidth_in_mbps
        }
        path = utils.append_uri(self._get_path(), id)
        if client_token is None:
            client_token = self._generate_default_client_token()
        params = {
            b'resize': b'',
            b'clientToken': client_token
        }
        return self._send_request(http_methods.PUT, path, params=params,
                                  body=json.dumps(body), config=config)

    @required(id=str)
    def get_eip_bp_detail(self, id, config=None):
        """
        get eip_bp's detail owned by the authenticated user and givened eip_bp_id.

        :type id: string
        :param eip: eip_bp's id.

        :type config: baidubce.BceClientConfiguration
        :param config:

        :return: detail of eip_bp, for example:
                {
                    "autoReleaseTime": "2020-05-30T06:46:44Z",
                    "name": "EIP_BP1588821183401",
                    "instanceId": "ip-9340430e",
                    "createTime": "2020-05-07T03:13:03Z",
                    "id": "bw-5fb3ce39",
                    "eips": [
                        "100.88.9.120"
                    ],
                    "instanceBandwidthInMbps": 1,
                    "bandwidthInMbps": 2,
                    "bindType": "eip"
                }
        """

        path = utils.append_uri(self._get_path(), id)
        return self._send_request(http_methods.GET, path, params=None,
                                  config=config)

    def list_eip_bps(self, id=None, name=None, bind_type=None,
                     marker=None, max_keys=1000, config=None):
        """
        get a list of eip_bp owned by the authenticated user and specified
        conditions. we can Also get a single eip_bp function  through this
        interface by eip_bp condition

        :type id: string
        :param id: eip_bp 's id conditions

        :type name: string
        :param name: eip_bp 's name condition

        :type bind_type: string
        :param bind_type: eip_bp 's bind_type condition, 'eip' or 'eipgroup'

        :type marker: string
        :param marker: The optional parameter marker specified in the original
         request to specify where in the results to begin listing.

        :type max_keys: int
        :param max_keys: The optional parameter to specifies the max number
         of list result to return. The default value is 1000.

        :type config: baidubce.BceClientConfiguration
        :param config:

        :return: list of eip_bp model, for example:
                {
                    "marker": "bw-5fb3ce39",
                    "maxKeys": 1000,
                    "nextMarker": null,
                    "bpList": [
                        {
                            "autoReleaseTime": "2020-05-30T06:46:44Z",
                            "name": "EIP_BP1588821183401",
                            "instanceId": "ip-9340430e",
                            "createTime": "2020-05-07T03:13:03Z",
                            "id": "bw-5fb3ce39",
                            "eips": [
                                "100.88.9.120"
                            ],
                            "bandwidthInMbps": 2,
                            "bindType": "eip"
                        }
                    ],
                    "isTruncated": false
                }
       """

        path = self._get_path()
        params = {}
        if id is not None:
            params[b'id'] = id
        if name is not None:
            params[b'name'] = name
        if bind_type is not None:
            params[b'bindType'] = bind_type
        if marker is not None:
            params[b'marker'] = marker
        if max_keys is not None:
            params[b'maxKeys'] = max_keys
        return self._send_request(http_methods.GET, path, params=params,
                                  config=config)

    @required(id=str, auto_release_time=str)
    def update_eip_bp_autoReleaseTime(self, id, auto_release_time, client_token=None,
                                      config=None):
        """
        Update eip_bp's auto_release_time

        :type id: string
        :param id: eip_bp's id

        :type auto_release_time: string (UTC format,like yyyy:mm:ddThh:mm:ssZ)
        :param auto_release_time: specify auto_release_time for eip_bp

        :type client_token: string
        :param client_token: if the clientToken is not specified by the user,
         a random string generated by default algorithm will be used.

        :type config: baidubce.BceClientConfiguration
        :param config:

        :return: BceResponse
        """
        body = {
            'autoReleaseTime': auto_release_time
        }
        path = utils.append_uri(self._get_path(), id)
        if client_token is None:
            client_token = self._generate_default_client_token()
        params = {
            b'retime': b'',
            b'clientToken': client_token
        }
        return self._send_request(http_methods.PUT, path, params=params,
                                  body=json.dumps(body), config=config)

    @required(id=str, name=str)
    def rename_eip_bp(self, id, name, client_token=None,
                      config=None):
        """
        Update eip_bp's name

        :type id: string
        :param id: eip_bp's id

        :type name: string
        :param name: eip_bp's name

        :type client_token: string
        :param client_token: if the clientToken is not specified by the user,
         a random string generated by default algorithm will be used.

        :type config: baidubce.BceClientConfiguration
        :param config:

        :return: BceResponse
        """
        body = {
            'name': name
        }
        path = utils.append_uri(self._get_path(), id)
        if client_token is None:
            client_token = self._generate_default_client_token()
        params = {
            b'rename': b'',
            b'clientToken': client_token
        }
        return self._send_request(http_methods.PUT, path, params=params,
                                  body=json.dumps(body), config=config)

    @required(id=str)
    def release_eip_bp(self, id, client_token=None, config=None):
        """
        release the eip_bp(delete operation)

        :type id: string
        :param id: eip_bp's id to be released

        :type client_token: string
        :param client_token: if the clientToken is not specified by the user,
         a random string generated by default algorithm will be used.

        :type config: baidubce.BceClientConfiguration
        :param config:

        :return: BceResponse
        """
        path = utils.append_uri(self._get_path(), id)
        if client_token is None:
            client_token = self._generate_default_client_token()
        params = {
            b'clientToken': client_token
        }
        return self._send_request(http_methods.DELETE, path, params=params,
                                  config=config)

    @staticmethod
    def _generate_default_client_token():
        """
        default client token by uuid1
        """
        return uuid.uuid1()

    @staticmethod
    def _get_path(prefix=None):
        """
        :type prefix: string
        :param prefix: path prefix
        """
        if prefix is None:
            prefix = EipBpClient.prefix
        return utils.append_uri(EipBpClient.version, prefix)

    def _merge_config(self, config):
        """

        :type config: baidubce.BceClientConfiguration
        :param config:
        :return:
        """
        if config is None:
            return self.config
        else:
            new_config = copy.copy(self.config)
            new_config.merge_non_none_values(config)
            return new_config

    def _send_request(self, http_method, path, body=None, headers=None,
                      params=None, config=None, body_parser=None):
        """

        :param http_method:
        :param path:
        :param body:
        :param headers:
        :param params:

        :type config: baidubce.BceClientConfiguration
        :param config:

        :param body_parser:

        :return: baidubce.BceResponse
        """
        config = self._merge_config(config)
        if body_parser is None:
            body_parser = handler.parse_json

        if headers is None:
            headers = {b'Accept': b'*/*',
                       b'Content-Type': b'application/json;charset=utf-8'}

        return bce_http_client.send_request(config, bce_v1_signer.sign,
                                            [handler.parse_error, body_parser],
                                            http_method, path, body, headers,
                                            params)



# Copyright (c) 2014 Baidu.com, Inc. All Rights Reserved
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions
# and limitations under the License.

"""
This module provides a client class for peer connection.
"""

import copy
import json
import logging
import uuid

from baidubce import utils
from baidubce import bce_base_client
from baidubce.auth import bce_v1_signer
from baidubce.http import bce_http_client
from baidubce.http import handler
from baidubce.http import http_methods
from baidubce.services.vpc import peerconn_model
from baidubce.utils import required

_logger = logging.getLogger(__name__)


default_billing_to_purchase_created = peerconn_model.Billing('Postpaid')
default_billing_to_purchase_reserved = peerconn_model.Billing()


class PeerConnClient(bce_base_client.BceBaseClient):
    """
    Peer connection sdk client
    """
    version = b'/v1'
    prefix = b'/peerconn'

    def __init__(self, config=None):
        bce_base_client.BceBaseClient.__init__(self, config)

    def _merge_config(self, config=None):
        """
        :param config:
        :type config: baidubce.BceClientConfiguration
        :return:
        """
        if config is None:
            return self.config
        else:
            new_config = copy.copy(self.config)
            new_config.merge_non_none_values(config)
            return new_config

    def _send_request(self, http_method, path,
                      body=None, headers=None, params=None,
                      config=None, body_parser=None):
        config = self._merge_config(config)
        if body_parser is None:
            body_parser = handler.parse_json
        if headers is None:
            headers = {b'Accept': b'*/*',
                       b'Content-Type': b'application/json;charset=utf-8'}
        return bce_http_client.send_request(
            config, bce_v1_signer.sign, [handler.parse_error, body_parser],
            http_method, path, body, headers, params)

    @required(bandwidth_in_mbps=int,
              local_vpc_id=(bytes, str),
              peer_vpc_id=(bytes, str),
              peer_region=(bytes, str))
    def create_peerconn(self, bandwidth_in_mbps, local_vpc_id,
                        peer_vpc_id, peer_region, description=None,
                        local_if_name=None, peer_account_id=None,
                        peer_if_name=None, client_token=None,
                        billing=None, config=None):
        """
        Create Peer connection.
        For peer connections within the same region, only postpaid is
        supported.
        For peer connections between different accounts, the peer connections
        are available only after the remote account accepts the connections.
        For peer connections within the same account, peer connection will be
        accepted automatically.
        There can be only one peer connection between any two VPCs.
        Peer connection endpoints cannot be the same VPC.
        If both local VPC and remote VPC are transit VPCs, peer connection
        cannot be established.

        :param client_token:
            An ASCII string whose length is less than 64.
            The request will be idempotent if client token is provided.
        :type client_token: string

        :param bandwidth_in_mbps:
            Network bandwidth (in unit of Mbps) of peer connection.
        :type bandwidth_in_mbps: int

        :param description:
            Description of peer connection.
        :type description: string

        :param local_if_name:
            Name of local interface of peer connection.
        :type local_if_name: string

        :param local_vpc_id:
            Local side VPC id of peer connection.
        :type local_vpc_id: string

        :param peer_account_id:
            Remote account id of peer connection.
            Used only when the peer connection connects two different accounts.
        :type peer_account_id: string

        :param peer_vpc_id:
            Remote side VPC id of peer connection.
        :type peer_vpc_id: string

        :param peer_region:
            Remote side region of peer connection.
        :type peer_region: string

        :param peer_if_name:
            Name of remote interface of peer connection.
        :type peer_if_name: string

        :param billing:
            Billing information.
        :type billing: nat_model.Billing

        :param config:
        :type config: baidubce.BceClientConfiguration

        :return:
        :rtype baidubce.bce_response.BceResponse
        """
        path = self._get_path()
        if client_token is None:
            client_token = generate_client_token()
        params = {
            b'clientToken': client_token
        }
        if billing is None:
            billing = default_billing_to_purchase_created
        body = {
            'bandwidthInMbps': bandwidth_in_mbps,
            'localVpcId': local_vpc_id,
            'peerVpcId': peer_vpc_id,
            'peerRegion': peer_region,
            'billing': billing.__dict__
        }
        if description is not None:
            body['description'] = description
        if local_if_name is not None:
            body['localIfName'] = local_if_name
        if peer_account_id is not None:
            body['peerAccountId'] = peer_account_id
        if peer_if_name is not None:
            body['peerIfName'] = peer_if_name
        return self._send_request(http_methods.POST,
                                  path, body=json.dumps(body),
                                  params=params, config=config)

    @required(vpc_id=(bytes, str))
    def list_peerconns(self, vpc_id, marker=None,
                       max_keys=None, config=None):
        """
        Return a list of peer connections.

        :param vpc_id:
            VPC id that peer connections connect to.
        :type vpc_id: string

        :param marker:
            The optional parameter marker specified in the original
            request to specify where in the results to begin listing.
            Together with the marker, specifies the list result which
            listing should begin. If the marker is not specified,
            the list result will listing from the first one.
        :type marker: string

        :param max_keys:
            The optional parameter to specifies the max number of
            list result to return.
            The default value is 1000.
        :type max_keys: int

        :param config:
        :type config: baidubce.BceClientConfiguration

        :return:
        :rtype baidubce.bce_response.BceResponse
        """
        path = self._get_path()
        params = {
            b'vpcId': vpc_id
        }
        if marker is not None:
            params[b'marker'] = marker
        if max_keys is not None:
            params[b'maxKeys'] = max_keys
        return self._send_request(http_methods.GET,
                                  path, params=params, config=config)

    @required(peer_conn_id=(bytes, str))
    def get_peerconn(self, peer_conn_id, config=None):
        """
        Get the detail information of specified peer connection.

        :param peer_conn_id:
            The id of specified peer connection.
        :type peer_conn_id: string

        :param config:
        :type config: baidubce.BceClientConfiguration

        :return:
        :rtype baidubce.bce_response.BceResponse
        """
        path = utils.append_uri(self._get_path(), peer_conn_id)
        return self._send_request(http_methods.GET, path, config=config)

    @required(peer_conn_id=(bytes, str), local_if_id=(bytes, str))
    def update_peerconn(self, peer_conn_id, local_if_id, description=None,
                        local_if_name=None, client_token=None, config=None):
        """
        Update the interface name or description of specified peer connection.

        :param peer_conn_id:
            The id of specified peer connection.
        :type peer_conn_id: string

        :param local_if_id:
            Local interface id of peer connection
        :type local_if_id: string

        :param description:
            Description of peer connection.
        :type description: string

        :param local_if_name:
            Name of local interface of peer connection.
        :type local_if_name: string

        :param client_token:
            An ASCII string whose length is less than 64.
            The request will be idempotent if clientToken is provided.
            If the clientToken is not specified by user,
            a random String generated by default algorithm will be used.
        :type client_token: string

        :param config:
        :type config: baidubce.BceClientConfiguration

        :return:
        :rtype baidubce.bce_response.BceResponse
        """
        path = utils.append_uri(self._get_path(), peer_conn_id)
        if client_token is None:
            client_token = generate_client_token()
        params = {
            b'clientToken': client_token
        }
        body = {
            'localIfId': local_if_id
        }
        if description is not None:
            body['description'] = description
        if local_if_name is not None:
            body['localIfName'] = local_if_name
        return self._send_request(http_methods.PUT,
                                  path, body=json.dumps(body),
                                  params=params, config=config)

    @required(peer_conn_id=(bytes, str), action=(bytes, str))
    def handle_peerconn(self, peer_conn_id, action, client_token=None,
                        config=None):
        """
        Accept or reject peer connection request.
        Timeout period of connection request is 7 days.
        When timeout or the remote side rejects the connection,
        the status of peer connection on initiator side is consulting failed.

        :param peer_conn_id:
            The id of specified peer connection.
        :type peer_conn_id: string

        :param action:
            'accept': when accepting the peer connection.
            'reject': when rejecting the peer connection.
        :type action: string

        :param client_token:
            An ASCII string whose length is less than 64.
            The request will be idempotent if clientToken is provided.
            If the clientToken is not specified by user,
            a random String generated by default algorithm will be used.
        :type client_token: string

        :param config:
        :type config: baidubce.BceClientConfiguration

        :return:
        :rtype baidubce.bce_response.BceResponse
        """
        path = utils.append_uri(self._get_path(), peer_conn_id)
        if client_token is None:
            client_token = generate_client_token()
        params = {
            b'clientToken': client_token
        }
        params[action] = None
        return self._send_request(http_methods.PUT,
                                  path, params=params, config=config)

    @required(peer_conn_id=(bytes, str))
    def delete_peerconn(self, peer_conn_id, client_token=None,
                        config=None):
        """
        Delete peer connection.
        For peer connections between different accounts,
        only initiator can perform this operation.
        One cannot delete prepaid peer connections that are not expired.
        Consulting failed prepaid peer connections can be deleted.

        :param peer_conn_id:
            The id of specified peer connection.
        :type peer_conn_id: string

        :param client_token:
            An ASCII string whose length is less than 64.
            The request will be idempotent if clientToken is provided.
            If the clientToken is not specified by user,
            a random String generated by default algorithm will be used.
        :type client_token: string

        :param config:
        :type config: baidubce.BceClientConfiguration

        :return:
        :rtype baidubce.bce_response.BceResponse
        """
        path = utils.append_uri(self._get_path(), peer_conn_id)
        if client_token is None:
            client_token = generate_client_token()
        params = {
            b'clientToken': client_token
        }
        return self._send_request(http_methods.DELETE,
                                  path, params=params, config=config)

    @required(peer_conn_id=(bytes, str), new_bandwidth_in_mbps=int)
    def resize_peerconn(self, peer_conn_id, new_bandwidth_in_mbps,
                        client_token=None, config=None):
        """
        Scale down/up the bandwidth of specified peer connection.
        For peer connections between different accounts,
        only initiator can perform this operation.
        Prepaid peer connection can only scale up.
        Postpaid peer connection can scale up or down.

        :param peer_conn_id:
            The id of specified peer connection.
        :type peer_conn_id: string

        :param client_token:
            An ASCII string whose length is less than 64.
            The request will be idempotent if clientToken is provided.
            If the clientToken is not specified by user,
            a random String generated by default algorithm will be used.
        :type client_token: string

        :param new_bandwidth_in_mbps:
            The new bandwidth of the peer connection.
        :type new_bandwidth_in_mbps: int

        :param config:
        :type config: baidubce.BceClientConfiguration

        :return:
        :rtype baidubce.bce_response.BceResponse
        """
        path = utils.append_uri(self._get_path(), peer_conn_id)
        if client_token is None:
            client_token = generate_client_token()
        params = {
            b'resize': None,
            b'clientToken': client_token
        }
        body = {
            'newBandwidthInMbps': new_bandwidth_in_mbps
        }
        return self._send_request(http_methods.PUT,
                                  path, body=json.dumps(body),
                                  params=params, config=config)

    @required(peer_conn_id=(bytes, str))
    def purchase_reserved_peerconn(self, peer_conn_id, client_token=None,
                                   billing=None, config=None):
        """
        Renew specified peer connection.
        Postpaid peer connection cannot be renewed.
        For peer connections between different accounts, only the initiator can
        perform this operation.

        :param peer_conn_id:
            The id of specified peer connection.
        :type peer_conn_id: string

        :param client_token:
            An ASCII string whose length is less than 64.
            The request will be idempotent if clientToken is provided.
            If the clientToken is not specified by user,
            a random String generated by default algorithm will be used.
        :type client_token: string

        :param billing:
            Billing information.
        :type billing: peerconn_model.Billing

        :param config:
        :type config: baidubce.BceClientConfiguration

        :return:
        :rtype baidubce.bce_response.BceResponse
        """
        path = utils.append_uri(self._get_path(), peer_conn_id)
        if client_token is None:
            client_token = generate_client_token()
        params = {
            b'purchaseReserved': None,
            b'clientToken': client_token
        }
        if billing is None:
            billing = default_billing_to_purchase_reserved
        body = {
            'billing': billing.__dict__
        }
        return self._send_request(http_methods.PUT,
                                  path, body=json.dumps(body),
                                  params=params, config=config)

    @required(peer_conn_id=(bytes, str), role=(bytes, str))
    def open_peerconn_dns_sync(self, peer_conn_id, role,
                               client_token=None, config=None):
        """
        Open DNS sync between VPCs connected by peer connection.
        DNS sync can be opened only when the status of
        peer connection is available.
        DNS sync cannot be opened when the DNS status of
        peer connection is syncing or closing.

        :param peer_conn_id:
            The id of specified peer connection.
        :type peer_conn_id: string

        :param role:
            'initiator': for VPC where peer connection is initiated.
            'acceptor': for VPC where peer connection is accepted.
        :type role: string

        :param client_token:
            An ASCII string whose length is less than 64.
            The request will be idempotent if clientToken is provided.
            If the clientToken is not specified by user,
            a random String generated by default algorithm will be used.
        :type client_token: string

        :param config:
        :type config: baidubce.BceClientConfiguration

        :return:
        :rtype baidubce.bce_response.BceResponse
        """
        path = utils.append_uri(self._get_path(), peer_conn_id)
        if client_token is None:
            client_token = generate_client_token()
        params = {
            b'open': None,
            b'role': role,
            b'clientToken': client_token
        }
        return self._send_request(http_methods.PUT,
                                  path, params=params, config=config)

    @required(peer_conn_id=(bytes, str), role=(bytes, str))
    def close_peerconn_dns_sync(self, peer_conn_id, role,
                                client_token=None, config=None):
        """
        Close DNS sync between VPCs connected by peer connection.
        DNS sync can be closed only when the status of peer connection
        is available. DNS sync cannot be closed when the DNS status
        of peer connection is syncing or closing.

        :param peer_conn_id:
            The id of specified peer connection.
        :type peer_conn_id: string

        :param role:
            'initiator': for VPC where peer connection is initiated.
            'acceptor': for VPC where peer connection is accepted.
        :type role: string

        :param client_token:
            An ASCII string whose length is less than 64.
            The request will be idempotent if clientToken is provided.
            If the clientToken is not specified by user,
            a random String generated by default algorithm will be used.
        :type client_token: string

        :param config:
        :type config: baidubce.BceClientConfiguration

        :return:
        :rtype baidubce.bce_response.BceResponse
        """
        path = utils.append_uri(self._get_path(), peer_conn_id)
        if client_token is None:
            client_token = generate_client_token()
        params = {
            b'close': None,
            b'role': role,
            b'clientToken': client_token
        }
        return self._send_request(http_methods.PUT,
                                  path, params=params, config=config)

    @staticmethod
    def _get_path(prefix=None):
        """
        :type prefix: string
        :param prefix: path prefix
        """
        if prefix is None:
            prefix = PeerConnClient.prefix
        return utils.append_uri(PeerConnClient.version, prefix)


def generate_client_token_by_uuid():
    """
    The default method to generate the random string for client_token
    if the optional parameter client_token is not specified by the user.

    :return:
    :rtype string
    """
    return str(uuid.uuid4())


generate_client_token = generate_client_token_by_uuid



# Copyright 2014 Baidu, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file
# except in compliance with the License. You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the
# License is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND,
# either express or implied. See the License for the specific language governing permissions
# and limitations under the License.

"""
This module provides a client class for BTS.
"""

import copy
import json
import logging

from baidubce.auth import bce_v1_signer
from baidubce.bce_base_client import BceBaseClient
from baidubce.exception import BceClientError
from baidubce.http import bce_http_client
from baidubce.http import handler
from baidubce.http import http_content_types
from baidubce.http import http_headers
from baidubce.http import http_methods
from baidubce.services import bts
from baidubce.services.bts import INVALID_ARGS_ERROR
from baidubce.services.bts.model import batch_query_row_args_2_dict
from baidubce.services.bts.model import create_instance_args_2_dict
from baidubce.services.bts.model import CreateInstanceArgs
from baidubce.services.bts.model import create_table_args_2_dict
from baidubce.services.bts.model import update_table_args_2_dict
from baidubce.services.bts.model import query_row_args_2_dict
from baidubce.services.bts.model import scan_args_2_dict
from baidubce.services.bts.util import _decode
from baidubce.services.bts.util import _encode

_logger = logging.getLogger(__name__)


class BtsClient(BceBaseClient):
    """
    BTS Client
    """
    def __init__(self, config=None):
        BceBaseClient.__init__(self, config)

    # ------------- instance operation -----------
    def create_instance(self, instance_name, create_instance_args=None, config=None):
        """
        create instance

        :param instance_name: instance name
        :type instance_name: string
        :param create_instance_args: arguments for create instance
        :type create_instance_args: CreateInstanceArgs
        :param config: None
        :type config: BceClientConfiguration

        :return:
        :rtype baidubce.bce_response.BceResponse
        """
        path = bts.URL_PREFIX + b"/" + instance_name
        if create_instance_args is None:
            create_instance_args = CreateInstanceArgs()
        return self._send_request(http_methods.PUT, path=path, config=config,
                                  body=json.dumps(create_instance_args, default=create_instance_args_2_dict),
                                  headers={http_headers.CONTENT_TYPE: http_content_types.JSON})

    def drop_instance(self, instance_name, config=None):
        """
        drop instance

        :param instance_name: instance name
        :type instance_name: string
        :param config: None
        :type config: BceClientConfiguration

        :return:
        :rtype baidubce.bce_response.BceResponse
        """
        path = bts.URL_PREFIX + b"/" + instance_name
        return self._send_request(http_methods.DELETE, path=path, config=config)

    def list_instances(self, config=None):
        """
        list instances

        :param config: None
        :type config: BceClientConfiguration

        :return:
        :rtype baidubce.bce_response.BceResponse
        """
        path = b"/v1/instances"
        return self._send_request(http_methods.GET, path=path, config=config)

    def show_instance(self, instance_name, config=None):
        """
        show instance

        :param instance_name: instance name
        :type instance_name: string
        :param config: None
        :type config: BceClientConfiguration

        :return:
        :rtype baidubce.bce_response.BceResponse
        """
        path = bts.URL_PREFIX + b"/" + instance_name
        return self._send_request(http_methods.GET, path=path, config=config)

    # ------------- table operation -----------
    def create_table(self, instance_name, table_name, create_table_args, config=None):
        """
        create table

        :param instance_name: instance name
        :type instance_name: string
        :param table_name: table name
        :type table_name: string
        :param create_table_args: arguments for create table
        :type create_table_args: CreateTableArgs
        :param config: None
        :type config: BceClientConfiguration

        :return:
        :rtype baidubce.bce_response.BceResponse
        """
        path = bts.URL_PREFIX + b"/" + instance_name + b"/table/" + table_name
        return self._send_request(http_methods.PUT, path=path, config=config,
                                  body=json.dumps(create_table_args, default=create_table_args_2_dict),
                                  headers={http_headers.CONTENT_TYPE: http_content_types.JSON})

    def update_table(self, instance_name, table_name, update_table_args, config=None):
        """
        update table

        :param instance_name: instance name
        :type instance_name: string
        :param table_name: table name
        :type table_name: string
        :param update_table_args: arguments for update table
        :type update_table_args: UpdateTableArgs
        :param config: None
        :type config: BceClientConfiguration

        :return:
        :rtype baidubce.bce_response.BceResponse
        """
        path = bts.URL_PREFIX + b"/" + instance_name + b"/table/" + table_name
        return self._send_request(http_methods.PUT, path=path, config=config,
                                  body=json.dumps(update_table_args, default=update_table_args_2_dict),
                                  headers={http_headers.CONTENT_TYPE: http_content_types.JSON})

    def drop_table(self, instance_name, table_name, config=None):
        """
        drop table

        :param instance_name: instance name
        :type instance_name: string
        :param table_name: table name
        :type table_name: string
        :param config: None
        :type config: BceClientConfiguration

        :return:
        :rtype baidubce.bce_response.BceResponse
        """
        path = bts.URL_PREFIX + b"/" + instance_name + b"/table/" + table_name
        return self._send_request(http_methods.DELETE, path=path, config=config)

    def show_table(self, instance_name, table_name, config=None):
        """
        show table

        :param instance_name: instance name
        :type instance_name: string
        :param table_name: table name
        :type table_name: string
        :param config: None
        :type config: BceClientConfiguration

        :return:
        :rtype baidubce.bce_response.BceResponse
        """
        path = bts.URL_PREFIX + b"/" + instance_name + b"/table/" + table_name
        return self._send_request(http_methods.GET, path=path, config=config)

    def list_tables(self, instance_name, config=None):
        """
        list tables

        :param instance_name: instance name
        :type instance_name: string
        :param config: None
        :type config: BceClientConfiguration

        :return:
        :rtype baidubce.bce_response.BceResponse
        """
        path = bts.URL_PREFIX + b"/" + instance_name + b"/tables"
        return self._send_request(http_methods.GET, path=path, config=config)

    # ------------- row operation -----------
    def put_row(self, instance_name, table_name, put_row_args, config=None):
        """
        put row

        :param instance_name: instance name
        :type instance_name: string
        :param table_name: table name
        :type table_name: string
        :param put_row_args: arguments for put row
        :type put_row_args: Row
        :param config: None
        :type config: BceClientConfiguration

        :return:
        :rtype baidubce.bce_response.BceResponse
        """
        if put_row_args is None or put_row_args.rowkey == "":
            ex = BceClientError(INVALID_ARGS_ERROR)
            _logger.debug(ex)
            raise ex
        try:
            put_row_args.rowkey = _encode(put_row_args.rowkey)
            for i in range(len(put_row_args.cells)):
                put_row_args.cells[i]["value"] = _encode(put_row_args.cells[i]["value"])
        except Exception as ex:
            raise ex

        path = bts.URL_PREFIX + b"/" + instance_name + b"/table/" + table_name + b"/row"
        return self._send_request(http_methods.PUT, path=path, config=config,
                                  body=json.dumps(put_row_args.__dict__),
                                  headers={http_headers.CONTENT_TYPE: http_content_types.JSON})

    def batch_put_row(self, instance_name, table_name, batch_put_row_args, config=None):
        """
        batch put row

        :param instance_name: instance name
        :type instance_name: string
        :param table_name: table name
        :type table_name: string
        :param batch_put_row_args: arguments for batch put row
        :type batch_put_row_args: BatchPutRowArgs
        :param config: None
        :type config: BceClientConfiguration

        :return:
        :rtype baidubce.bce_response.BceResponse
        """
        if batch_put_row_args is None:
            ex = BceClientError(INVALID_ARGS_ERROR)
            _logger.debug(ex)
            raise ex
        try:
            for i in range(len(batch_put_row_args.rows)):
                if batch_put_row_args.rows[i]["rowkey"] == "":
                    ex = BceClientError(INVALID_ARGS_ERROR)
                    _logger.debug(ex)
                    raise ex
                batch_put_row_args.rows[i]["rowkey"] = _encode(batch_put_row_args.rows[i]["rowkey"])
                for j in range(len(batch_put_row_args.rows[i]["cells"])):
                    batch_put_row_args.rows[i]["cells"][j]["value"] = \
                        _encode(batch_put_row_args.rows[i]["cells"][j]["value"])
        except Exception as ex:
            raise ex

        path = bts.URL_PREFIX + b"/" + instance_name + b"/table/" + table_name + b"/rows"
        return self._send_request(http_methods.PUT, path=path, config=config,
                                  body=json.dumps(batch_put_row_args.__dict__),
                                  headers={http_headers.CONTENT_TYPE: http_content_types.JSON})

    def delete_row(self, instance_name, table_name, delete_row_args, config=None):
        """
        delete row

        :param instance_name: instance name
        :type instance_name: string
        :param table_name: table name
        :type table_name: string
        :param delete_row_args: arguments for delete row
        :type delete_row_args: QueryRowArgs
        :param config: None
        :type config: BceClientConfiguration

        :return:
        :rtype baidubce.bce_response.BceResponse
        """
        if delete_row_args is None or delete_row_args.rowkey == "":
            ex = BceClientError(INVALID_ARGS_ERROR)
            _logger.debug(ex)
            raise ex
        try:
            delete_row_args.rowkey = _encode(delete_row_args.rowkey)
        except Exception as ex:
            raise ex

        path = bts.URL_PREFIX + b"/" + instance_name + b"/table/" + table_name + b"/row"
        return self._send_request(http_methods.DELETE, path=path, config=config,
                                  body=json.dumps(delete_row_args, default=query_row_args_2_dict),
                                  headers={http_headers.CONTENT_TYPE: http_content_types.JSON})

    def batch_delete_row(self, instance_name, table_name, batch_delete_row_args, config=None):
        """
        batch delete row

        :param instance_name: instance name
        :type instance_name: string
        :param table_name: table name
        :type table_name: string
        :param batch_delete_row_args: arguments for batch delete row
        :type batch_delete_row_args: BatchQueryRowArgs
        :param config: None
        :type config: BceClientConfiguration

        :return:
        :rtype baidubce.bce_response.BceResponse
        """
        if batch_delete_row_args is None:
            ex = BceClientError(INVALID_ARGS_ERROR)
            _logger.debug(ex)
            raise ex
        try:
            for i in range(len(batch_delete_row_args.rows)):
                if batch_delete_row_args.rows[i]["rowkey"] == "":
                    ex = BceClientError(INVALID_ARGS_ERROR)
                    _logger.debug(ex)
                    raise ex
                batch_delete_row_args.rows[i]["rowkey"] = _encode(batch_delete_row_args.rows[i]["rowkey"])
        except Exception as ex:
            raise ex

        path = bts.URL_PREFIX + b"/" + instance_name + b"/table/" + table_name + b"/rows"
        return self._send_request(http_methods.DELETE, path=path, config=config,
                                  body=json.dumps(batch_delete_row_args, default=batch_query_row_args_2_dict),
                                  headers={http_headers.CONTENT_TYPE: http_content_types.JSON})

    def get_row(self, instance_name, table_name, get_row_args, config=None):
        """
        get row

        :param instance_name: instance name
        :type instance_name: string
        :param table_name: table name
        :type table_name: string
        :param get_row_args: arguments for get row
        :type get_row_args: QueryRowArg
        :param config: None
        :type config: BceClientConfiguration

        :return:
        :rtype baidubce.bce_response.BceResponse
        """
        try:
            if get_row_args is None or get_row_args.rowkey == "":
                ex = BceClientError(INVALID_ARGS_ERROR)
                _logger.debug(ex)
                raise ex
            get_row_args.rowkey = _encode(get_row_args.rowkey)
        except Exception as ex:
            raise ex

        path = bts.URL_PREFIX + b"/" + instance_name + b"/table/" + table_name + b"/row"
        response = self._send_request(http_methods.GET, path=path, config=config,
                                      body=json.dumps(get_row_args, default=query_row_args_2_dict),
                                      headers={http_headers.CONTENT_TYPE: http_content_types.JSON})
        try:
            if response.result is not None:
                response.result[0].rowkey = _decode(str(response.result[0].rowkey))
                for i in range(len(response.result[0].cells)):
                    response.result[0].cells[i].value = _decode(str(response.result[0].cells[i].value))
        except Exception as ex:
            raise ex
        return response

    def batch_get_row(self, instance_name, table_name, batch_get_row_args, config=None):
        """
        batch get row

        :param instance_name: instance name
        :type instance_name: string
        :param table_name: table name
        :type table_name: string
        :param batch_get_row_args: arguments for batch get row
        :type batch_get_row_args: BatchQueryRowArgs
        :param config: None
        :type config: BceClientConfiguration

        :return:
        :rtype baidubce.bce_response.BceResponse
        """
        if batch_get_row_args is None:
            ex = BceClientError(INVALID_ARGS_ERROR)
            _logger.debug(ex)
            raise ex
        try:
            for i in range(len(batch_get_row_args.rows)):
                if batch_get_row_args.rows[i]["rowkey"] == "":
                    ex = BceClientError(INVALID_ARGS_ERROR)
                    _logger.debug(ex)
                    raise ex
                batch_get_row_args.rows[i]["rowkey"] = _encode(batch_get_row_args.rows[i]["rowkey"])
        except Exception as ex:
            raise ex

        path = bts.URL_PREFIX + b"/" + instance_name + b"/table/" + table_name + b"/rows"
        response = self._send_request(http_methods.GET, path=path, config=config,
                                      body=json.dumps(batch_get_row_args, default=batch_query_row_args_2_dict),
                                      headers={http_headers.CONTENT_TYPE: http_content_types.JSON})
        try:
            if response.result is not None:
                for i in range(len(response.result)):
                    response.result[i].rowkey = _decode(str(response.result[i].rowkey))
                    for j in range(len(response.result[i].cells)):
                        response.result[i].cells[j].value = _decode(str(response.result[i].cells[j].value))
        except Exception as ex:
            raise ex
        return response

    def scan(self, instance_name, table_name, scan_args, config=None):
        """
        scan

        :param instance_name: instance name
        :type instance_name: string
        :param table_name: table name
        :type table_name: string
        :param scan_args: arguments for scan
        :type scan_args: ScanArgs
        :param config: None
        :type config: BceClientConfiguration

        :return:
        :rtype baidubce.bce_response.BceResponse
        """
        if scan_args is None:
            ex = BceClientError(INVALID_ARGS_ERROR)
            _logger.debug(ex)
            raise ex
        try:
            if scan_args.start_rowkey is not "":
                scan_args.start_rowkey = _encode(scan_args.start_rowkey)
            if scan_args.stop_rowkey is not "":
                scan_args.stop_rowkey = _encode(scan_args.stop_rowkey)
        except Exception as ex:
            raise ex

        path = bts.URL_PREFIX + b"/" + instance_name + b"/table/" + table_name + b"/rows"
        response = self._send_request(http_methods.GET, path=path, config=config,
                                      body=json.dumps(scan_args, default=scan_args_2_dict),
                                      headers={http_headers.CONTENT_TYPE: http_content_types.JSON})
        try:
            if response.result is not None:
                for i in range(len(response.result)):
                    response.result[i].rowkey = _decode(str(response.result[i].rowkey))
                    for j in range(len(response.result[i].cells)):
                        response.result[i].cells[j].value = _decode(str(response.result[i].cells[j].value))
        except Exception as ex:
            raise ex
        return response

    def _merge_config(self, config):
        if config is None:
            return self.config
        else:
            new_config = copy.copy(self.config)
            new_config.merge_non_none_values(config)
            return new_config

    # ------------- Http Request -----------
    def _send_request(
            self, http_method, path,
            body=None, headers=None, params=None,
            config=None,
            body_parser=None):
        config = self._merge_config(config)
        if body_parser is None:
            body_parser = handler.parse_json

        if config.security_token is not None:
            headers = headers or {}
            headers[http_headers.STS_SECURITY_TOKEN] = config.security_token

        return bce_http_client.send_request(
            config, bce_v1_signer.sign, [handler.parse_error, body_parser],
            http_method, path, body, headers, params)


